import decimal
import os
from contextlib import closing
from pathlib import Path
from typing import AsyncGenerator, Generator, Literal, cast

import numpy
import pyarrow as pa
import pytest
import pytest_asyncio

from tiled._tests.utils import temp_postgres
from tiled.adapters.sql import (
    arrow_schema_to_column_defns,
    arrow_schema_to_create_table,
)
from tiled.storage import EmbeddedSQLStorage, RemoteSQLStorage


@pytest_asyncio.fixture
async def postgresql_uri() -> AsyncGenerator[str, None]:
    uri = os.getenv("TILED_TEST_POSTGRESQL_URI")
    if uri is None:
        pytest.skip("TILED_TEST_POSTGRESQL_URI is not set")

    async with temp_postgres(uri) as uri_with_database_name:
        yield uri_with_database_name
        # yield uri_with_database_name.rsplit("/", 1)[0]


@pytest_asyncio.fixture
def sqlite_uri(tmp_path: Path) -> Generator[str, None, None]:
    yield f"sqlite:///{tmp_path}/test.db"


@pytest_asyncio.fixture
def duckdb_uri(tmp_path: Path) -> Generator[str, None, None]:
    yield f"duckdb:///{tmp_path}/test.db"


INT8_INFO = numpy.iinfo(numpy.int8)
INT16_INFO = numpy.iinfo(numpy.int16)
INT32_INFO = numpy.iinfo(numpy.int32)
INT64_INFO = numpy.iinfo(numpy.int64)
UINT8_INFO = numpy.iinfo(numpy.uint8)
UINT16_INFO = numpy.iinfo(numpy.uint16)
UINT32_INFO = numpy.iinfo(numpy.uint32)
UINT64_INFO = numpy.iinfo(numpy.uint64)
FLOAT16_INFO = numpy.finfo(numpy.float16)
FLOAT32_INFO = numpy.finfo(numpy.float32)
FLOAT64_INFO = numpy.finfo(numpy.float64)
# Map schemas (testing different data types or combinations of data types)
# to an inner mapping. The inner mapping maps each dialect to a tuple,
# (SQL type definition, Arrow type read back).
# { test_case_id:  (input_table, {dialect: (expected_typedefs, expected_schema)})}
TEST_CASES = {
    "bool": (
        pa.Table.from_arrays([pa.array([True, False], "bool")], names=["x"]),
        {
            "duckdb": (["BOOLEAN NULL"], pa.schema([("x", "bool")])),
            "sqlite": (["INTEGER NULL"], pa.schema([("x", "int64")])),
            "postgresql": (["BOOLEAN NULL"], pa.schema([("x", "bool")])),
        },
    ),
    "string": (
        pa.Table.from_arrays([pa.array(["a", "b"], "string")], names=["x"]),
        {
            "duckdb": (["VARCHAR NULL"], pa.schema([("x", "string")])),
            "sqlite": (["TEXT NULL"], pa.schema([("x", "string")])),
            "postgresql": (["TEXT NULL"], pa.schema([("x", "string")])),
        },
    ),
    "int8": (
        pa.Table.from_arrays(
            [pa.array([INT8_INFO.min, INT8_INFO.max], "int8")], names=["x"]
        ),
        {
            "duckdb": (["TINYINT NULL"], pa.schema([("x", "int8")])),
            "sqlite": (["INTEGER NULL"], pa.schema([("x", "int64")])),
            "postgresql": (["SMALLINT NULL"], pa.schema([("x", "int16")])),
        },
    ),
    "int16": (
        pa.Table.from_arrays(
            [pa.array([INT16_INFO.min, INT16_INFO.max], "int16")], names=["x"]
        ),
        {
            "duckdb": (["SMALLINT NULL"], pa.schema([("x", "int16")])),
            "sqlite": (["INTEGER NULL"], pa.schema([("x", "int64")])),
            "postgresql": (["SMALLINT NULL"], pa.schema([("x", "int16")])),
        },
    ),
    "int32": (
        pa.Table.from_arrays(
            [pa.array([INT32_INFO.min, INT32_INFO.max], "int32")], names=["x"]
        ),
        {
            "duckdb": (["INTEGER NULL"], pa.schema([("x", "int32")])),
            "sqlite": (["INTEGER NULL"], pa.schema([("x", "int64")])),
            "postgresql": (["INTEGER NULL"], pa.schema([("x", "int32")])),
        },
    ),
    "int64": (
        pa.Table.from_arrays(
            [pa.array([INT64_INFO.min, INT64_INFO.max], "int64")], names=["x"]
        ),
        {
            "duckdb": (["BIGINT NULL"], pa.schema([("x", "int64")])),
            "sqlite": (["INTEGER NULL"], pa.schema([("x", "int64")])),
            "postgresql": (["BIGINT NULL"], pa.schema([("x", "int64")])),
        },
    ),
    "uint8": (
        pa.Table.from_arrays(
            [pa.array([UINT8_INFO.min, UINT8_INFO.max], "uint8")], names=["x"]
        ),
        {
            "duckdb": (["UTINYINT NULL"], pa.schema([("x", "uint8")])),
            "sqlite": (["INTEGER NULL"], pa.schema([("x", "int64")])),
            "postgresql": (["SMALLINT NULL"], pa.schema([("x", "int16")])),
        },
    ),
    "uint16": (
        pa.Table.from_arrays(
            [pa.array([UINT16_INFO.min, UINT16_INFO.max], "uint16")], names=["x"]
        ),
        {
            "duckdb": (["USMALLINT NULL"], pa.schema([("x", "uint16")])),
            "sqlite": (["INTEGER NULL"], pa.schema([("x", "int64")])),
            "postgresql": (["INTEGER NULL"], pa.schema([("x", "int32")])),
        },
    ),
    "uint32": (
        pa.Table.from_arrays(
            [pa.array([UINT32_INFO.min, UINT32_INFO.max], "uint32")], names=["x"]
        ),
        {
            "duckdb": (["UINTEGER NULL"], pa.schema([("x", "uint32")])),
            "sqlite": (["INTEGER NULL"], pa.schema([("x", "int64")])),
            "postgresql": (["BIGINT NULL"], pa.schema([("x", "int64")])),
        },
    ),
    "uint64": (
        pa.Table.from_arrays(
            [pa.array([UINT64_INFO.min, UINT64_INFO.max], "uint64")], names=["x"]
        ),
        {
            "duckdb": (["UBIGINT NULL"], pa.schema([("x", "uint64")])),
        },
    ),
    "list_of_ints": (
        pa.Table.from_arrays(
            [pa.array([[1, 2], [3, 4]], pa.list_(pa.int32()))], names=["x"]
        ),
        {
            "duckdb": (["INTEGER[] NULL"], pa.schema([("x", pa.list_(pa.int32()))])),
            "postgresql": (
                ["INTEGER ARRAY NULL"],
                pa.schema([("x", pa.list_(pa.int32()))]),
            ),
        },
    ),
    "list_of_bounded_ints": (
        pa.Table.from_arrays(
            [pa.array([[1, 2], [3, 4]], pa.list_(pa.int32(), 2))], names=["x"]
        ),
        {
            "duckdb": (["INTEGER[] NULL"], pa.schema([("x", pa.list_(pa.int32()))])),
            "postgresql": (
                ["INTEGER ARRAY NULL"],
                pa.schema([("x", pa.list_(pa.int32()))]),
            ),
        },
    ),
    "float16": (
        pa.Table.from_arrays(
            [pa.array([FLOAT16_INFO.min, FLOAT16_INFO.max], "float16")], names=["x"]
        ),
        {},  # not supported by any backend
    ),
    "float32": (
        pa.Table.from_arrays(
            [pa.array([FLOAT32_INFO.min, FLOAT32_INFO.max], "float32")], names=["x"]
        ),
        {
            "duckdb": (["REAL NULL"], pa.schema([("x", "float32")])),
            "sqlite": (["REAL NULL"], pa.schema([("x", "double")])),
            "postgresql": (["REAL NULL"], pa.schema([("x", "float32")])),
        },
    ),
    "float64": (
        pa.Table.from_arrays(
            [pa.array([FLOAT64_INFO.min, FLOAT64_INFO.max], "float64")], names=["x"]
        ),
        {
            "duckdb": (["DOUBLE NULL"], pa.schema([("x", "float64")])),
            "sqlite": (["REAL NULL"], pa.schema([("x", "double")])),
            "postgresql": (["DOUBLE PRECISION NULL"], pa.schema([("x", "float64")])),
        },
    ),
    "decimal": (
        pa.Table.from_arrays(
            [pa.array([decimal.Decimal("123.45")], pa.decimal128(5, 2))], names=["x"]
        ),
        {
            "duckdb": (["DECIMAL(5, 2) NULL"], pa.schema([("x", pa.decimal128(5, 2))])),
        },
    ),
}


@pytest.mark.parametrize("dialect", ["duckdb", "postgresql", "sqlite"])
@pytest.mark.parametrize("test_case_id", list(TEST_CASES))
def test_data_types(
    test_case_id: str,
    dialect: Literal["postgresql", "sqlite", "duckdb"],
    request: pytest.FixtureRequest,
) -> None:
    test_table_name = f"test_{test_case_id}"
    table, dialect_results = TEST_CASES[test_case_id]

    if (dialect == "duckdb") and (test_case_id == "decimal"):
        pytest.xfail(reason="Regression in support, needs investigation")

    if dialect not in cast(dict, dialect_results):  # type: ignore
        with pytest.raises(ValueError, match="Unsupported PyArrow type"):
            arrow_schema_to_column_defns(table.schema, dialect)
        return

    expected_typedefs, expected_schema = dialect_results[dialect]  # type: ignore
    db_uri = request.getfixturevalue(f"{dialect}_uri")
    columns = arrow_schema_to_column_defns(table.schema, dialect)
    assert list(columns.values()) == expected_typedefs

    query = arrow_schema_to_create_table(table.schema, test_table_name, dialect)

    storage_cls = RemoteSQLStorage if dialect == "postgresql" else EmbeddedSQLStorage
    storage = storage_cls(uri=db_uri)
    with closing(storage.connect()) as conn:
        with conn.cursor() as cursor:
            cursor.execute(query)
        conn.commit()

        # For SQLite specifically, some inference is needed by ADBC to get the type
        # and on an empty table the value is not defined. As of this writing it is
        # int64 by default; in the future it may be null.
        # https://github.com/apache/arrow-adbc/issues/581
        if dialect != "sqlite":
            assert conn.adbc_get_table_schema(test_table_name) == expected_schema

        with conn.cursor() as cursor:
            cursor.adbc_ingest(test_table_name, table, mode="append")

        assert conn.adbc_get_table_schema(test_table_name) == expected_schema

        with conn.cursor() as cursor:
            cursor.execute(f"SELECT * FROM {test_table_name}")
            result = cursor.fetch_arrow_table()

        # The result will match expected_schema, which may not be the same as
        # the schema the data was uploaded as, if the databases does not support
        # that precise type.
        assert result.schema == expected_schema

        # Before comparing the Tables, we cast the Table into the original schema,
        # which might use finer types.
        assert result.cast(table.schema) == table

    storage.dispose()  # Close all connections before deleting the storage DB



# This tests an experimental feature likely to be deprecated.
# https://github.com/bluesky/tiled/issues/217
import pathlib

import pytest

from tiled.catalog import in_memory
from tiled.client import Context, from_context
from tiled.server.app import build_app


@pytest.fixture(scope="module")
def module_tmp_path(tmp_path_factory: pytest.TempdirFactory) -> pathlib.Path:
    return tmp_path_factory.mktemp("temp")


@pytest.fixture(scope="module")
def client(module_tmp_path):
    catalog = in_memory(writable_storage=str(module_tmp_path))
    app = build_app(catalog)
    with Context.from_app(app) as context:
        client = from_context(context)
        client.write_array([1], key="x", metadata={"sample": {"color": "red"}})
        client.write_array([2], key="y", metadata={"sample": {"color": "blue"}})
        yield client


def test_select_metadata(client):
    http_client = client.context.http_client
    # /metadata
    response = http_client.get("/api/v1/metadata/x?select_metadata=[sample.color]")
    result = response.json()
    assert result["data"]["attributes"]["metadata"] == {"selected": ["red"]}
    # /search
    response = http_client.get("/api/v1/search/?select_metadata=[sample.color]")
    result = response.json()
    for item, color in zip(result["data"], ["red", "blue"]):
        assert item["attributes"]["metadata"] == {"selected": [color]}



from pathlib import Path

import numpy
import yaml

from tiled._tests.utils import enter_username_password
from tiled.client import Context, from_context
from tiled.server.app import build_app_from_config

CONFIG_NAME = "toy_authentication.yml"
CATALOG_STORAGE = "data/"


def main():
    file_directory = Path(__file__).resolve().parent
    config_directory = file_directory.parent
    Path(file_directory, CATALOG_STORAGE).mkdir()

    with open(Path(config_directory, CONFIG_NAME)) as config_file:
        config = yaml.load(config_file, Loader=yaml.BaseLoader)
    app = build_app_from_config(config)
    context = Context.from_app(app)
    with enter_username_password("admin", "admin"):
        client = from_context(context, remember_me=False)
    for n in ["A", "B", "C", "D"]:
        client.write_array(
            key=n, array=10 * numpy.ones((10, 10)), access_tags=[f"data_{n}"]
        )
    client.logout()
    context.close()


if __name__ == "__main__":
    main()



"Download utilities implemented using httpx and rich progress bars, with parallelism."
import re
import signal
from concurrent.futures import ThreadPoolExecutor, wait
from pathlib import Path
from threading import Event
from typing import Iterable

import httpx
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskID,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

from .utils import handle_error, retry_context

# This extracts the filename to the Content-Disposition header.
CONTENT_DISPOSITION_PATTERN = re.compile(r"^attachment; ?filename=\"(.*)\"$")

# This is used by the caller of download(...) as a placeholder
# that should be substituted with the filename provided by the
# server in the Content-Disposition header.
ATTACHMENT_FILENAME_PLACEHOLDER = "CONTENT_DISPOSITION_HEADER_ATTACHMENT_FILENAME"


def _download_url(
    progress: Progress,
    task_id: TaskID,
    done_event: Event,
    client: httpx.Client,
    url: str,
    path: str,
) -> None:
    """Copy data from a url to a local file."""
    progress.console.log(f"Requesting {url}")
    try:
        path.parent.mkdir(exist_ok=True, parents=True)
        for attempt in retry_context():
            with attempt:
                with handle_error(client.stream("GET", url)) as response:
                    if path.name == ATTACHMENT_FILENAME_PLACEHOLDER:
                        # Use filename provided by server.
                        filename = CONTENT_DISPOSITION_PATTERN.match(
                            response.headers["Content-Disposition"]
                        ).group(1)
                        path = Path(path.parent, filename)
                    with open(path, "wb") as file:
                        total = int(response.headers["Content-Length"])
                        progress.update(task_id, total=total)
                        progress.start_task(task_id)
                        for chunk in response.iter_bytes():
                            file.write(chunk)
                            progress.update(task_id, advance=len(chunk))
                            if done_event.is_set():
                                return
    except Exception as err:
        progress.console.log(f"ERROR {err!r}")
    else:
        progress.console.log(f"Downloaded {path}")
    return path


def download(
    client, urls: Iterable[str], paths: Iterable[Path], *, max_workers: int = 4
):
    """
    Download multiple URLs to given paths, in parallel.
    """
    progress = Progress(
        # TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
        BarColumn(bar_width=None),
        "[progress.percentage]{task.percentage:>3.1f}%",
        "•",
        DownloadColumn(),
        "•",
        TransferSpeedColumn(),
        "•",
        TimeRemainingColumn(),
    )

    if len(urls) != len(paths):
        raise ValueError(
            "Must provide a list of URLs and a list of paths with equal length. "
            f"Received {len(urls)=} and {len(paths)=}."
        )

    def sigint_handler(signum, frame):
        done_event.set()
        original_sigint_handler(signal.SIGINT, frame)

    done_event = Event()
    original_sigint_handler = signal.getsignal(signal.SIGINT)
    signal.signal(signal.SIGINT, sigint_handler)
    futures = []
    try:
        with progress:
            with ThreadPoolExecutor(max_workers=max_workers) as pool:
                for url, path in zip(urls, paths):
                    task_id = progress.add_task("download", start=False)
                    future = pool.submit(
                        _download_url, progress, task_id, done_event, client, url, path
                    )
                    futures.append(future)
                wait(futures)
    finally:
        # Restore SIGINT handler.
        signal.signal(signal.SIGINT, original_sigint_handler)
    return [future.result() for future in futures]



import json
import logging

import numpy as np
import pyarrow

from locust import HttpUser, between, events, task
from tiled.client import from_uri


@events.init_command_line_parser.add_listener
def _(parser):
    parser.add_argument(
        "--api-key",
        type=str,
        default="secret",
        help="API key for Tiled authentication (default: secret)",
    )
    parser.add_argument(
        "--container-name",
        type=str,
        default="locust_testing",
        help="Container name for test data (default: locust_testing)",
    )


@events.init.add_listener
def on_locust_init(environment, **kwargs):
    if environment.host is None:
        raise ValueError(
            "Host must be specified with --host argument, or through the web-ui."
        )

    environment.container_name = environment.parsed_options.container_name
    environment.known_dataset_key = create_test_dataset(
        environment.host, environment.parsed_options.api_key, environment.container_name
    )


def create_test_dataset(host, api_key, container_name):
    """Create a test dataset using Tiled client for reading tasks"""

    # Connect to Tiled server using client
    root_client = from_uri(host, api_key=api_key)

    # Create container if it doesn't exist
    if container_name not in root_client:
        root_client.create_container(container_name)
    client = root_client[container_name]

    rng = np.random.default_rng(seed=42)
    rng.integers(10, size=100, dtype=np.dtype("uint8"))

    # Write and read tabular data to the SQL storage
    table = pyarrow.Table.from_pydict({"a": rng.random(100), "b": rng.random(100)})

    table_client = client.create_appendable_table(table.schema)
    table_client.append_partition(table, 0)

    # Verify we can read it back
    result = table_client.read()
    logging.debug(f"Created and verified dataset: {result}")

    dataset_id = table_client.item["id"]
    client.logout()
    return dataset_id


class ReadingUser(HttpUser):
    """User that reads data from Tiled using HTTP API"""

    wait_time = between(0.5, 2)

    def on_start(self):
        self.client.headers = {
            "Authorization": f"Apikey {self.environment.parsed_options.api_key}"
        }

    @task(1)
    def read_table_data(self):
        """Read table data from our known dataset"""
        # Read the table data we created
        self.client.get(
            f"/api/v1/table/full/{self.environment.container_name}/{self.environment.known_dataset_key}",
        )

    @task(1)
    def read_metadata(self):
        """Read metadata from our known dataset"""
        self.client.get(
            f"/api/v1/metadata/{self.environment.container_name}/{self.environment.known_dataset_key}",
        )

    @task(1)
    def root_endpoint(self):
        """Test root endpoint performance"""
        self.client.get("/")

    @task(1)
    def metadata_root(self):
        """Test metadata root endpoint"""
        self.client.get("/api/v1/metadata/")

    @task(1)
    def read_table_partition(self):
        """Read specific partition from our known dataset"""
        url = (
            f"/api/v1/table/partition/{self.environment.container_name}/"
            f"{self.environment.known_dataset_key}?partition=0"
        )
        self.client.get(url)

    @task(1)
    def healthz_endpoint(self):
        """Test health check endpoint"""
        self.client.get("/healthz")

    @task(1)
    def metrics_endpoint(self):
        """Test Prometheus metrics endpoint"""
        self.client.get("/api/v1/metrics")

    @task(1)
    def about_endpoint(self):
        """Test API information endpoint"""
        self.client.get("/api/v1/")

    @task(1)
    def search_root(self):
        """Test search at root level"""
        self.client.get("/api/v1/search/")

    @task(1)
    def container_full_endpoint(self):
        """Test container full data endpoint"""
        self.client.get(f"/api/v1/container/full/{self.environment.container_name}")

    @task(1)
    def whoami_endpoint(self):
        """Test user identity endpoint"""
        self.client.get("/api/v1/auth/whoami")

    @task(1)
    def search_fulltext(self):
        """Test fulltext search queries"""
        params = {"filter[fulltext][condition][text]": "test"}
        self.client.get("/api/v1/search/", params=params)

    @task(1)
    def search_with_limit(self):
        """Test search with limit parameter"""
        params = {"page[limit]": 5}
        self.client.get("/api/v1/search/", params=params)

    @task(1)
    def search_with_pagination(self):
        """Test search with offset and limit parameters"""
        params = {"page[offset]": 10, "page[limit]": 5}
        self.client.get("/api/v1/search/", params=params)

    @task(1)
    def search_with_sort(self):
        """Test search with sort parameter"""
        params = {"sort": "key"}
        self.client.get("/api/v1/search/", params=params)

    @task(1)
    def search_with_max_depth(self):
        """Test search with max_depth parameter"""
        params = {"max_depth": 1}
        self.client.get("/api/v1/search/", params=params)

    @task(1)
    def search_structure_family(self):
        """Test structure family search queries"""
        params = {"filter[structure_family][condition][value]": ["table"]}
        self.client.get("/api/v1/search/", params=params)

    @task(1)
    def search_with_omit_links(self):
        """Test search with omit_links parameter"""
        params = {"omit_links": "true"}
        self.client.get("/api/v1/search/", params=params)

    @task(1)
    def search_with_data_sources(self):
        """Test search with include_data_sources parameter"""
        params = {"include_data_sources": "true"}
        self.client.get("/api/v1/search/", params=params)

    @task(1)
    def search_eq(self):
        """Test equality search queries"""
        params = {
            "filter[eq][condition][key]": "structure_family",
            "filter[eq][condition][value]": json.dumps("table"),
        }
        self.client.get("/api/v1/search/", params=params)

    @task(1)
    def search_noteq(self):
        """Test not equal search queries"""
        params = {
            "filter[noteq][condition][key]": "structure_family",
            "filter[noteq][condition][value]": json.dumps("array"),
        }
        self.client.get("/api/v1/search/", params=params)

    @task(1)
    def search_comparison(self):
        """Test comparison search queries"""
        params = {
            "filter[comparison][condition][operator]": "gt",
            "filter[comparison][condition][key]": "id",
            "filter[comparison][condition][value]": json.dumps(0),
        }
        self.client.get("/api/v1/search/", params=params)

    @task(1)
    def search_like(self):
        """Test like search queries"""
        params = {
            "filter[like][condition][key]": "key",
            "filter[like][condition][pattern]": json.dumps("%"),
        }
        self.client.get("/api/v1/search/", params=params)

    @task(1)
    def search_contains(self):
        """Test contains search queries"""
        params = {
            "filter[contains][condition][key]": "key",
            "filter[contains][condition][value]": json.dumps("locust"),
        }
        self.client.get("/api/v1/search/", params=params)

    @task(1)
    def search_in(self):
        """Test in search queries"""
        params = {
            "filter[in][condition][key]": "structure_family",
            "filter[in][condition][value]": '["table", "array"]',
        }
        self.client.get("/api/v1/search/", params=params)

    @task(1)
    def search_notin(self):
        """Test not in search queries"""
        params = {
            "filter[notin][condition][key]": "structure_family",
            "filter[notin][condition][value]": '["sparse"]',
        }
        self.client.get("/api/v1/search/", params=params)



from collections.abc import Mapping
from pathlib import Path
from typing import Any, Iterator


class DirectoryContainer(Mapping[str, bytes]):
    """ """

    def __init__(self, directory: Path, form: Any):
        """

        Parameters
        ----------
        directory :
        form :
        """
        self.directory = directory
        self.form = form

    def __getitem__(self, form_key: str) -> bytes:
        """

        Parameters
        ----------
        form_key :

        Returns
        -------

        """
        with open(self.directory / form_key, "rb") as file:
            return file.read()

    def __setitem__(self, form_key: str, value: bytes) -> None:
        """

        Parameters
        ----------
        form_key :
        value :

        Returns
        -------

        """
        with open(self.directory / form_key, "wb") as file:
            file.write(value)

    def __delitem__(self, form_key: str) -> None:
        """

        Parameters
        ----------
        form_key :

        Returns
        -------

        """
        (self.directory / form_key).unlink(missing_ok=True)

    def __iter__(self) -> Iterator[str]:
        """

        Returns
        -------

        """
        yield from self.form.expected_from_buffers()

    def __len__(self) -> int:
        """

        Returns
        -------

        """
        return len(self.form.expected_from_buffers())



from __future__ import annotations

import enum
import uuid
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import pydantic.generics
from pydantic import ConfigDict, Field, StringConstraints
from pydantic_core import PydanticCustomError
from typing_extensions import Annotated, TypedDict

from ..structures.array import ArrayStructure
from ..structures.awkward import AwkwardStructure
from ..structures.core import STRUCTURE_TYPES, Spec, StructureFamily
from ..structures.data_source import Management
from ..structures.sparse import SparseStructure
from ..structures.table import TableStructure

if TYPE_CHECKING:
    import tiled.authn_database.orm
    import tiled.catalog.orm

DataT = TypeVar("DataT")
LinksT = TypeVar("LinksT")
MetaT = TypeVar("MetaT")
StructureT = TypeVar("StructureT")


MAX_ALLOWED_SPECS = 20


class Error(pydantic.BaseModel):
    code: int
    message: str


class Response(pydantic.BaseModel, Generic[DataT, LinksT, MetaT]):
    data: Optional[DataT]
    error: Optional[Error] = None
    links: Optional[LinksT] = None
    meta: Optional[MetaT] = None

    @pydantic.field_validator("error")
    def check_consistency(cls, v, values):
        if v is not None and values["data"] is not None:
            raise ValueError("must not provide both data and error")
        if v is None and values.get("data") is None:
            raise ValueError("must provide data or error")
        return v


class PaginationLinks(pydantic.BaseModel):
    self: str
    next: str
    prev: str
    first: str
    last: str


class EntryFields(str, enum.Enum):
    metadata = "metadata"
    structure_family = "structure_family"
    structure = "structure"
    count = "count"
    sorting = "sorting"
    specs = "specs"
    data_sources = "data_sources"
    none = ""
    access_blob = "access_blob"


class NodeStructure(pydantic.BaseModel):
    contents: Optional[Dict[str, Any]]
    count: int

    model_config = pydantic.ConfigDict(extra="forbid")


class SortingDirection(int, enum.Enum):
    ASCENDING = 1
    DESCENDING = -1


class SortingItem(pydantic.BaseModel):
    key: str
    direction: SortingDirection


# class Spec(pydantic.BaseModel, extra="forbid", frozen=True):
#     name: Annotated[str, StringConstraints(max_length=255)]
#     version: Optional[Annotated[str, StringConstraints(max_length=255)]] = None


# Wait for fix https://github.com/pydantic/pydantic/issues/3957
# Specs = pydantic.conlist(Spec, max_length=20, unique_items=True)
Specs = Annotated[List[Spec], Field(max_length=MAX_ALLOWED_SPECS)]


class Asset(pydantic.BaseModel):
    data_uri: str
    is_directory: bool
    parameter: Optional[str] = None
    num: Optional[int] = None
    id: Optional[int] = None

    @classmethod
    def from_orm(cls, orm: tiled.catalog.orm.Asset) -> Asset:
        return cls(
            data_uri=orm.data_uri,
            is_directory=orm.is_directory,
            id=orm.id,
        )

    @classmethod
    def from_assoc_orm(cls, orm):
        return cls(
            data_uri=orm.asset.data_uri,
            is_directory=orm.asset.is_directory,
            parameter=orm.parameter,
            num=orm.num,
            id=orm.asset.id,
        )


class Revision(pydantic.BaseModel):
    revision_number: int
    metadata: dict
    specs: Specs
    access_blob: dict
    time_updated: datetime

    @classmethod
    def from_orm(cls, orm: tiled.catalog.orm.Revision) -> Revision:
        # Trailing underscore in 'metadata_' avoids collision with
        # SQLAlchemy reserved word 'metadata'.
        return cls(
            revision_number=orm.revision_number,
            metadata=orm.metadata_,
            specs=orm.specs,
            access_blob=orm.access_blob,
            time_updated=orm.time_updated,
        )


class Patch(pydantic.BaseModel):
    offset: Tuple[int, ...]
    shape: Tuple[int, ...]
    extend: bool


class DataSource(pydantic.BaseModel, Generic[StructureT]):
    id: Optional[int] = None
    structure_family: StructureFamily
    structure: Optional[StructureT]
    mimetype: Optional[str] = None
    parameters: dict = {}
    assets: List[Asset] = []
    management: Management = Management.writable

    model_config = pydantic.ConfigDict(extra="forbid")

    @classmethod
    def from_orm(cls, orm: tiled.catalog.orm.DataSource) -> DataSource:
        if hasattr(orm.structure, "structure"):
            structure_cls = STRUCTURE_TYPES[orm.structure_family]
            structure = structure_cls.from_json(orm.structure.structure)
        else:
            structure = None
        return cls(
            id=orm.id,
            structure_family=orm.structure_family,
            structure=structure,
            mimetype=orm.mimetype,
            parameters=orm.parameters,
            assets=[Asset.from_assoc_orm(assoc) for assoc in orm.asset_associations],
            management=orm.management,
        )


class NodeAttributes(pydantic.BaseModel):
    ancestors: List[str]
    structure_family: Optional[StructureFamily] = None
    specs: Optional[Specs] = None
    metadata: Optional[Dict] = None  # free-form, user-specified dict
    structure: Optional[
        Union[
            ArrayStructure,
            AwkwardStructure,
            SparseStructure,
            NodeStructure,
            TableStructure,
        ]
    ] = None
    access_blob: Optional[Dict] = None  # free-form, access_policy-specified dict

    sorting: Optional[List[SortingItem]] = None
    data_sources: Optional[List[DataSource]] = None

    model_config = pydantic.ConfigDict(extra="forbid")


AttributesT = TypeVar("AttributesT")
ResourceMetaT = TypeVar("ResourceMetaT")
ResourceLinksT = TypeVar("ResourceLinksT")


class SelfLinkOnly(pydantic.BaseModel):
    self: str


class ContainerLinks(pydantic.BaseModel):
    self: str
    search: str
    full: str


class ArrayLinks(pydantic.BaseModel):
    self: str
    full: str
    block: str


class AwkwardLinks(pydantic.BaseModel):
    self: str
    buffers: str
    full: str


class DataFrameLinks(pydantic.BaseModel):
    self: str
    full: str
    partition: str


class SparseLinks(pydantic.BaseModel):
    self: str
    full: str
    block: str


resource_links_type_by_structure_family = {
    StructureFamily.array: ArrayLinks,
    StructureFamily.awkward: AwkwardLinks,
    StructureFamily.container: ContainerLinks,
    StructureFamily.sparse: SparseLinks,
    StructureFamily.table: DataFrameLinks,
}


class EmptyDict(pydantic.BaseModel):
    pass


class ContainerMeta(pydantic.BaseModel):
    count: int


class Resource(pydantic.BaseModel, Generic[AttributesT, ResourceLinksT, ResourceMetaT]):
    "A JSON API Resource"
    id: Union[str, uuid.UUID]
    attributes: AttributesT
    links: Optional[ResourceLinksT] = None
    meta: Optional[ResourceMetaT] = None


class AccessAndRefreshTokens(pydantic.BaseModel):
    access_token: str
    expires_in: int
    refresh_token: str
    refresh_token_expires_in: int
    token_type: str


class RefreshToken(pydantic.BaseModel):
    refresh_token: str


class DeviceCode(pydantic.BaseModel):
    device_code: str
    grant_type: str


class PrincipalType(str, enum.Enum):
    user = "user"
    service = "service"


class Identity(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(from_attributes=True)
    id: Annotated[str, StringConstraints(max_length=255)]
    provider: Annotated[str, StringConstraints(max_length=255)]
    latest_login: Optional[datetime] = None

    @classmethod
    def from_orm(cls, orm: tiled.authn_database.orm.Identity) -> Identity:
        return cls(id=orm.id, provider=orm.provider, latest_login=orm.latest_login)


class Role(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(from_attributes=True)
    name: str
    scopes: List[str]
    # principals

    @classmethod
    def from_orm(cls, orm: tiled.authn_database.orm.Role) -> Role:
        return cls(name=orm.name, scopes=orm.scopes)


class APIKey(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(from_attributes=True)
    first_eight: Annotated[str, StringConstraints(min_length=8, max_length=8)]
    expiration_time: Optional[datetime] = None
    note: Optional[Annotated[str, StringConstraints(max_length=255)]] = None
    scopes: List[str]
    access_tags: Optional[List[str]] = None
    latest_activity: Optional[datetime] = None

    @classmethod
    def from_orm(cls, orm: tiled.authn_database.orm.APIKey) -> APIKey:
        return cls(
            first_eight=orm.first_eight,
            expiration_time=orm.expiration_time,
            note=orm.note,
            scopes=orm.scopes,
            access_tags=orm.access_tags,
            latest_activity=orm.latest_activity,
        )


class APIKeyWithSecret(APIKey):
    secret: str  # hex-encoded bytes

    @classmethod
    def from_orm(
        cls, orm: tiled.authn_database.orm.APIKeyWithSecret, secret: str
    ) -> APIKeyWithSecret:
        return cls(
            first_eight=orm.first_eight,
            expiration_time=orm.expiration_time,
            note=orm.note,
            scopes=orm.scopes,
            access_tags=orm.access_tags,
            latest_activity=orm.latest_activity,
            secret=secret,
        )


class Session(pydantic.BaseModel):
    """
    This related to refresh tokens, which have a session uuid ("sid") claim.

    When the client attempts to use a refresh token, we first check
    here to ensure that the "session", which is associated with a chain
    of refresh tokens that came from a single authentication, are still valid.
    """

    # The id field (primary key) is intentionally not exposed to the application.
    # It is left as an internal database concern.
    model_config = pydantic.ConfigDict(from_attributes=True)
    uuid: uuid.UUID
    expiration_time: datetime
    revoked: bool
    state: Optional[Dict[Any, Any]]

    @classmethod
    def from_orm(cls, orm: tiled.authn_database.orm.Session) -> Session:
        return cls(
            uuid=orm.uuid,
            expiration_time=orm.expiration_time,
            revoked=orm.revoked,
            state=orm.state,
        )


class Principal(pydantic.BaseModel):
    "Represents a User or Service"
    # The id field (primary key) is intentionally not exposed to the application.
    # It is left as an internal database concern.
    model_config = pydantic.ConfigDict(from_attributes=True)
    uuid: uuid.UUID
    type: PrincipalType
    identities: List[Identity] = []
    roles: List[Role] = []
    api_keys: List[APIKey] = []
    sessions: List[Session] = []
    latest_activity: Optional[datetime] = None

    @classmethod
    def from_orm(
        cls,
        orm: tiled.authn_database.orm.Principal,
        latest_activity: Optional[datetime] = None,
    ) -> Principal:
        return cls(
            uuid=orm.uuid,
            type=orm.type,
            identities=[Identity.from_orm(id_) for id_ in orm.identities],
            roles=[Role.from_orm(id_) for id_ in orm.roles],
            api_keys=[APIKey.from_orm(api_key) for api_key in orm.api_keys],
            sessions=[Session.from_orm(session) for session in orm.sessions],
            latest_activity=latest_activity,
        )


class APIKeyRequestParams(pydantic.BaseModel):
    # Provide an example for expires_in. Otherwise, OpenAPI suggests lifetime=0.
    # If the user is not reading carefully, they will be frustrated when they
    # try to use the instantly-expiring API key!
    expires_in: Optional[int] = pydantic.Field(
        ..., json_schema_extra={"example": 600}
    )  # seconds
    access_tags: Optional[List[str]] = pydantic.Field(
        default=None, json_schema_extra={"example": ["writing_tag", "public"]}
    )
    scopes: Optional[List[str]] = pydantic.Field(
        ..., json_schema_extra={"example": ["inherit"]}
    )
    note: Optional[str] = None


class PostMetadataRequest(pydantic.BaseModel):
    id: Optional[str] = None
    structure_family: StructureFamily
    metadata: Dict = {}
    data_sources: List[DataSource] = []
    specs: Specs = []
    access_blob: Optional[Dict] = {}

    # Wait for fix https://github.com/pydantic/pydantic/issues/3957
    # to do this with `unique_items` parameters to `pydantic.constr`.
    @pydantic.field_validator("specs")
    def specs_uniqueness_validator(cls, v):
        if v is None:
            return None
        for i, value in enumerate(v, start=1):
            if value in v[i:]:
                raise ValueError
        return v

    @pydantic.model_validator(mode="after")
    def narrow_structure_type(self):
        "Convert the structure on each data_source from a dict to the appropriate pydantic model."
        for data_source in self.data_sources:
            if self.structure_family != StructureFamily.container:
                structure_cls = STRUCTURE_TYPES[self.structure_family]
                if data_source.structure is not None:
                    data_source.structure = structure_cls.from_json(
                        data_source.structure
                    )
        return self


class PutDataSourceRequest(pydantic.BaseModel):
    data_source: DataSource
    patch: Optional[Patch]


class PostMetadataResponse(pydantic.BaseModel, Generic[ResourceLinksT]):
    id: str
    links: Union[ArrayLinks, DataFrameLinks, SparseLinks]
    metadata: Dict
    data_sources: List[DataSource]
    access_blob: Dict


class PutMetadataResponse(pydantic.BaseModel, Generic[ResourceLinksT]):
    id: str
    links: Union[ArrayLinks, DataFrameLinks, SparseLinks]
    # May be None if not altered
    metadata: Optional[Dict] = None
    data_sources: Optional[List[DataSource]] = None
    access_blob: Optional[Dict] = None


class DistinctValueInfo(pydantic.BaseModel):
    value: Any = None
    count: Optional[int] = None


class GetDistinctResponse(pydantic.BaseModel):
    metadata: Optional[Dict[str, List[DistinctValueInfo]]] = None
    structure_families: Optional[List[DistinctValueInfo]] = None
    specs: Optional[List[DistinctValueInfo]] = None


class PutMetadataRequest(pydantic.BaseModel):
    # These fields are optional because None means "no changes; do not update".
    metadata: Optional[Dict] = None
    specs: Optional[Specs] = None
    access_blob: Optional[Dict] = None

    # Wait for fix https://github.com/pydantic/pydantic/issues/3957
    # to do this with `unique_items` parameters to `pydantic.constr`.
    @pydantic.field_validator("specs")
    def specs_uniqueness_validator(cls, v):
        if v is None:
            return None
        for i, value in enumerate(v, start=1):
            if value in v[i:]:
                raise ValueError
        return v


def JSONPatchType(dtype=Any):
    # we use functional syntax with TypedDict here since "from" is a keyword
    return TypedDict(
        "JSONPatchType",
        {
            "op": str,
            "path": str,
            "from": str,
            "value": dtype,
        },
        total=False,
    )


JSONPatchSpec = JSONPatchType(Spec)
JSONPatchAny = JSONPatchType(Any)


class HyphenizedBaseModel(pydantic.BaseModel):
    # This model configuration allows aliases like "content-type"
    model_config = ConfigDict(alias_generator=lambda f: f.replace("_", "-"))


class PatchMetadataRequest(HyphenizedBaseModel):
    content_type: str

    # These fields are optional because None means "no changes; do not update".
    # Dict for merge-patch:
    metadata: Optional[Union[List[JSONPatchAny], Dict]] = None

    # Specs for merge-patch. left_to_right mode is used to distinguish between
    # merge-patch List[asdict(Spec)] and json-patch List[Dict]
    specs: Optional[Union[Specs, List[JSONPatchSpec]]] = Field(
        union_mode="left_to_right"
    )

    # These fields are optional because None means "no changes; do not update".
    # Dict for merge-patch:
    # Define an alias to override parent class alias generator
    access_blob: Optional[Union[List[JSONPatchAny], Dict]] = Field(
        alias="access_blob", default=None
    )

    @pydantic.field_validator("specs")
    def specs_uniqueness_validator(cls, v):
        if v is None:
            return None
        if v and isinstance(v[0], Spec):
            # This is the MERGE_PATCH case
            if len(v) != len(set(v)):
                raise PydanticCustomError("specs", "Items must be unique")
        elif v and isinstance(v[0], dict):
            # This is the JSON_PATCH case
            v_new = [v_["value"] for v_ in v if v_["op"] in ["add", "replace"]]
            # Note: uniqueness should be checked with existing specs included,
            # however since we use replace_metadata to eventually write to db this
            # will be caught and an error raised there.
            if len(v_new) != len(set(v_new)):
                raise PydanticCustomError("specs", "Items must be unique")
        return v


class PatchMetadataResponse(pydantic.BaseModel, Generic[ResourceLinksT]):
    id: str
    links: Union[ArrayLinks, DataFrameLinks, SparseLinks]
    # May be None if not altered
    metadata: Optional[Dict]
    data_sources: Optional[List[DataSource]]
    access_blob: Optional[Dict]


SearchResponse = Response[
    List[Resource[NodeAttributes, Dict, Dict]], PaginationLinks, Dict
]


class EnvelopeFormat(str, enum.Enum):
    json = "json"
    msgpack = "msgpack"


NodeStructure.model_rebuild()



"""Add 'composite' structure family to the structurefamily enum.

Revision ID: 45a702586b2a
Revises: ed3a4223a600
Create Date: 2025-03-12 15:05:25.335010

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "45a702586b2a"
down_revision = "ed3a4223a600"
branch_labels = None
depends_on = None


def upgrade():
    connection = op.get_bind()

    if connection.engine.dialect.name == "postgresql":
        with op.get_context().autocommit_block():
            op.execute(
                sa.text(
                    "ALTER TYPE structurefamily ADD VALUE IF NOT EXISTS 'composite' AFTER 'awkward'"
                )
            )


def downgrade():
    # PostgreSQL does not support dropping values from enums.
    # (The enum would need to be destroyed and recreated, migrating
    # the data in the process.) Fortunately, an extra unused values does
    # not interfere with operation of older versions of Tiled.
    pass



import asyncio
import collections
import copy
import dataclasses
import importlib
import itertools as it
import logging
import operator
import os
import re
import shutil
import sys
import uuid
from contextlib import closing
from datetime import datetime
from functools import partial, reduce
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union, cast
from urllib.parse import urlparse

import anyio
import orjson
from fastapi import HTTPException, WebSocketDisconnect
from sqlalchemy import (
    and_,
    delete,
    event,
    exists,
    false,
    func,
    literal,
    not_,
    or_,
    select,
    text,
    true,
    type_coerce,
    update,
)
from sqlalchemy.dialects.postgresql import ARRAY, JSONB, REGCONFIG, TEXT
from sqlalchemy.engine import make_url
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, create_async_engine
from sqlalchemy.orm import aliased, selectinload
from sqlalchemy.pool import AsyncAdaptedQueuePool
from sqlalchemy.sql.expression import cast as sql_cast
from sqlalchemy.sql.sqltypes import MatchType
from starlette.status import HTTP_404_NOT_FOUND, HTTP_415_UNSUPPORTED_MEDIA_TYPE

from tiled.queries import (
    AccessBlobFilter,
    Comparison,
    Contains,
    Eq,
    FullText,
    In,
    KeyPresent,
    KeysFilter,
    Like,
    NotEq,
    NotIn,
    Operator,
    SpecsQuery,
    StructureFamilyQuery,
)

from ..mimetypes import (
    APACHE_ARROW_FILE_MIME_TYPE,
    AWKWARD_BUFFERS_MIMETYPE,
    DEFAULT_ADAPTERS_BY_MIMETYPE,
    PARQUET_MIMETYPE,
    SPARSE_BLOCKS_PARQUET_MIMETYPE,
    TILED_SQL_TABLE_MIMETYPE,
    ZARR_MIMETYPE,
)
from ..query_registration import QueryTranslationRegistry
from ..server.core import NoEntry
from ..server.schemas import Asset, DataSource, Management, Revision
from ..storage import (
    FileStorage,
    SQLStorage,
    get_storage,
    parse_storage,
    register_storage,
)
from ..structures.core import Spec, StructureFamily
from ..utils import (
    UNCHANGED,
    Conflicts,
    OneShotCachedMap,
    UnsupportedQueryType,
    ensure_awaitable,
    ensure_specified_sql_driver,
    import_object,
    path_from_uri,
    safe_json_dump,
)
from . import orm
from .core import check_catalog_database, initialize_database
from .explain import ExplainAsyncSession
from .utils import compute_structure_id

logger = logging.getLogger(__name__)

DEFAULT_ECHO = bool(int(os.getenv("TILED_ECHO_SQL", "0") or "0"))
INDEX_PATTERN = re.compile(r"^[A-Za-z0-9_-]+$")

# When data is uploaded, how is it saved?
# TODO: Make this configurable at Catalog construction time.
DEFAULT_CREATION_MIMETYPE = {
    StructureFamily.array: ZARR_MIMETYPE,
    StructureFamily.awkward: AWKWARD_BUFFERS_MIMETYPE,
    StructureFamily.table: PARQUET_MIMETYPE,
    StructureFamily.sparse: SPARSE_BLOCKS_PARQUET_MIMETYPE,
}

# TODO: make type[Adapter] after #1047
STORAGE_ADAPTERS_BY_MIMETYPE = OneShotCachedMap[str, type](
    {
        ZARR_MIMETYPE: lambda: importlib.import_module(
            "...adapters.zarr", __name__
        ).ZarrArrayAdapter,
        AWKWARD_BUFFERS_MIMETYPE: lambda: importlib.import_module(
            "...adapters.awkward_buffers", __name__
        ).AwkwardBuffersAdapter,
        PARQUET_MIMETYPE: lambda: importlib.import_module(
            "...adapters.parquet", __name__
        ).ParquetDatasetAdapter,
        "text/csv": lambda: importlib.import_module(
            "...adapters.csv", __name__
        ).CSVAdapter,
        SPARSE_BLOCKS_PARQUET_MIMETYPE: lambda: importlib.import_module(
            "...adapters.sparse_blocks_parquet", __name__
        ).SparseBlocksParquetAdapter,
        APACHE_ARROW_FILE_MIME_TYPE: lambda: importlib.import_module(
            "...adapters.arrow", __name__
        ).ArrowAdapter,
        TILED_SQL_TABLE_MIMETYPE: lambda: importlib.import_module(
            "...adapters.sql", __name__
        ).SQLAdapter,
    }
)


class RootNode:
    """
    Node representing the root of the tree.

    This node is special because the state configuring this node arises from
    server initialization (typically a configuration file) not from the
    database.

    It mocks the relevant part of the interface of .orm.Node.
    """

    structure_family = StructureFamily.container

    def __init__(self, metadata, specs, top_level_access_blob):
        self.id = 0
        self.parent = None
        self.metadata_ = metadata or {}
        self.specs = specs or []
        self.key = ""
        self.data_sources = None
        self.access_blob = top_level_access_blob or {}


class Context:
    def __init__(
        self,
        engine: AsyncEngine,
        writable_storage=None,
        readable_storage=None,
        adapters_by_mimetype=None,
        cache_settings=None,
        key_maker=lambda: str(uuid.uuid4()),
        storage_pool_size=None,
        storage_max_overflow=None,
    ):
        self.engine = engine

        self.writable_storage = []
        self.readable_storage = set()

        # Back-compat: `writable_storage` used to be a dict: we want its values.
        if isinstance(writable_storage, dict):
            writable_storage = list(writable_storage.values())
        # Back-compat: `writable_storage` used to be a filepath.
        if isinstance(writable_storage, (str, Path)):
            writable_storage = [writable_storage]
        if isinstance(readable_storage, str):
            raise ValueError("readable_storage should be a list of URIs or paths")

        for item in writable_storage or []:
            self.writable_storage.append(
                parse_storage(
                    item, pool_size=storage_pool_size, max_overflow=storage_max_overflow
                )
            )
        for item in readable_storage or []:
            self.readable_storage.add(parse_storage(item))
        # Writable storage should also be readable.
        self.readable_storage.update(self.writable_storage)
        # Register all storage in a registry that enables Adapters to access
        # credentials (if applicable).
        for item in self.readable_storage:
            register_storage(item)

        self.key_maker = key_maker
        adapters_by_mimetype = adapters_by_mimetype or {}
        # If adapters_by_mimetype comes from a configuration file,
        # objects are given as importable strings, like "package.module:Reader".
        for key, value in list(adapters_by_mimetype.items()):
            if isinstance(value, str):
                adapters_by_mimetype[key] = import_object(value)
        # User-provided adapters take precedence over defaults.
        merged_adapters_by_mimetype = collections.ChainMap(
            adapters_by_mimetype, DEFAULT_ADAPTERS_BY_MIMETYPE
        )
        self.adapters_by_mimetype = merged_adapters_by_mimetype
        self.cache_settings = cache_settings

    def session(self):
        "Convenience method for constructing an AsyncSession context"
        return ExplainAsyncSession(self.engine, autoflush=False, expire_on_commit=False)

    async def execute(self, statement, explain=None):
        "Debugging convenience utility, not exposed to server"
        async with self.session() as db:
            result = await db.execute(text(statement), explain=explain)
            await db.commit()
            return result

    async def startup(self):
        if (self.engine.dialect.name == "sqlite") and (
            self.engine.url.database == ":memory:"
            or self.engine.url.query.get("mode") == "memory"
        ):
            # Special-case for in-memory SQLite: Because it is transient we can
            # skip over anything related to migrations.
            await initialize_database(self.engine)
        else:
            await check_catalog_database(self.engine)

        cache_client = None
        cache_ttl = 0
        if self.cache_settings:
            if self.cache_settings["uri"].startswith("redis"):
                from redis import asyncio as redis

                socket_timeout = self.cache_settings.get("socket_timeout", 10.0)
                socket_connect_timeout = self.cache_settings.get(
                    "socket_connect_timeout", 10.0
                )
                cache_client = redis.from_url(
                    self.cache_settings["uri"],
                    socket_timeout=socket_timeout,
                    socket_connect_timeout=socket_connect_timeout,
                )
                cache_ttl = self.cache_settings.get("ttl", 3600)
        self.cache_client = cache_client
        self.cache_ttl = cache_ttl


class CatalogNodeAdapter:
    query_registry = QueryTranslationRegistry()
    register_query = query_registry.register
    register_query_lazy = query_registry.register_lazy

    def __init__(
        self,
        context,
        node,
        *,
        conditions=None,
        queries=None,
        sorting=None,
        mount_path: Optional[list[str]] = None,
    ):
        self.context = context
        self.node = node
        self.sorting = sorting or [("", 1)]
        self.order_by_clauses = order_by_clauses(self.sorting)
        self.conditions = conditions or []
        self.queries = queries or []
        self.structure_family = node.structure_family
        self.specs = [Spec(**spec) for spec in node.specs]
        self.startup_tasks = [self.startup]
        if mount_path:
            self.startup_tasks.append(partial(self.create_mount, mount_path))
        self.shutdown_tasks = [self.shutdown]

    async def path_segments(self):
        statement = (
            select(orm.Node.key)
            .where(orm.Node.id != 0)
            .join(orm.NodesClosure, orm.NodesClosure.ancestor == orm.Node.id)
            .where(orm.NodesClosure.descendant == self.node.id)
            .order_by(orm.NodesClosure.depth.desc())
        )

        async with self.context.session() as db:
            return (await db.execute(statement)).scalars().all()

    @property
    def access_blob(self):
        return self.node.access_blob

    def metadata(self):
        return self.node.metadata_

    async def startup(self):
        await self.context.startup()

    async def create_mount(self, mount_path: list[str]):
        statement = node_from_segments(mount_path).with_only_columns(orm.Node.id)
        async with self.context.engine.connect() as conn:
            self.node.id = (await conn.execute(statement)).scalar()
        self.node.key = mount_path[-1]

    async def shutdown(self):
        await self.context.engine.dispose()

    @property
    def writable(self):
        return bool(self.context.writable_storage)

    @property
    def key(self):
        return self.node.key

    def __repr__(self):
        return f"<{type(self).__name__} {self.key}>"

    async def __aiter__(self):
        statement = select(orm.Node.key).filter(orm.Node.parent == self.node.id)
        for condition in self.conditions:
            statement = statement.filter(condition)
        async with self.context.session() as db:
            return (
                (await db.execute(statement.order_by(*self.order_by_clauses)))
                .scalars()
                .all()
            )
        statement = select(orm.Node.key).filter(orm.Node.parent == self.node.id)
        async with self.context.session() as db:
            return (await db.execute(statement)).scalar().all()

    @property
    def data_sources(self):
        return [DataSource.from_orm(ds) for ds in (self.node.data_sources or [])]

    async def asset_by_id(self, asset_id):
        statement = (
            select(orm.Asset)
            .join(
                orm.DataSourceAssetAssociation,
                orm.DataSourceAssetAssociation.asset_id == orm.Asset.id,
            )
            .join(
                orm.DataSource,
                orm.DataSource.id == orm.DataSourceAssetAssociation.data_source_id,
            )
            .join(orm.Node, orm.Node.id == orm.DataSource.node_id)
            .where((orm.Node.id == self.node.id) & (orm.Asset.id == asset_id))
        )
        async with self.context.session() as db:
            asset = (await db.execute(statement)).scalar()
        if asset is None:
            return None  # no match
        return Asset.from_orm(asset)

    def structure(self):
        if self.data_sources:
            assert len(self.data_sources) == 1  # more not yet implemented
            return self.data_sources[0].structure
        return None

    def apply_conditions(self, statement):
        # IF this is a sqlite database and we are doing a full text MATCH
        # query, we need a JOIN with the FTS5 virtual table.
        if (self.context.engine.dialect.name == "sqlite") and any(
            isinstance(condition.type, MatchType) for condition in self.conditions
        ):
            statement = statement.join(
                orm.metadata_fts5, orm.metadata_fts5.c.rowid == orm.Node.id
            )
        for condition in self.conditions:
            statement = statement.filter(condition)
        return statement

    async def exact_len(self):
        "Get the exact number of child nodes."
        statement = (
            select(func.count())
            .select_from(orm.Node)
            .filter(orm.Node.parent == self.node.id)
        )
        statement = self.apply_conditions(statement)

        async with self.context.session() as db:
            return (await db.execute(statement)).scalar_one()

    async def approx_len(self) -> Optional[int]:
        """Get an approximate number of child nodes using table statistics.

        This only works for PostgreSQL databases and does not take into account
        any filtering conditions. To be able to use these queries, the `nodes`
        must be vacuumed and analyzed regularly (at least once).

        If the database is not PostgreSQL, or if the statistics can not be
        obtained, return None.
        """

        if self.context.engine.dialect.name == "postgresql":
            async with self.context.session() as db:
                parent_and_freqs = await db.execute(
                    text(
                        """
                SELECT unnest(most_common_vals::text::int[])::int AS parent,
                       unnest(most_common_freqs) AS freq
                FROM pg_stats
                WHERE schemaname = 'public' AND tablename = 'nodes' AND attname = 'parent';
                                """
                    )
                )
                for parent, freq in parent_and_freqs:
                    if parent == self.node.id:
                        total = (
                            await db.execute(
                                text(
                                    """
                            SELECT reltuples::bigint FROM pg_class
                            WHERE  oid = 'public.nodes'::regclass;
                                            """
                                )
                            )
                        ).scalar_one()
                        return int(total * freq)
                else:
                    return None  # Statistics can not be obtained

        elif self.context.engine.dialect.name == "sqlite":
            # SQLite has no statistics tables, so we fall back to exact count.
            return None

    async def lbound_len(self, threshold) -> int:
        """Get a fast lower bound on the number of child nodes.

        This only counts up to `threshold`+1 nodes, so is fast even for large
        containers. If result is <= `threshold`, it is exact.
        """

        limited = (
            select(literal(1))
            .select_from(orm.Node)
            .where(orm.Node.parent == self.node.id)
            .limit(threshold + 1)
        )
        limited = self.apply_conditions(limited).cte("limited")
        statement = select(func.count()).select_from(limited)

        async with self.context.session() as db:
            return (await db.execute(statement)).scalar_one()

    async def lookup_adapter(self, segments: list[str]):
        # TODO: Accept filter for predicate-pushdown.
        if not segments:
            return self
        if self.conditions and len(segments) > 1:
            # There are some conditions (i.e. WHERE clauses) applied to
            # this node, either via user search queries or via access
            # control policy queries. Look up first the _direct_ child of this
            # node, if it exists within the filtered results.
            first_level = await self.lookup_adapter(segments[:1])

            # Now proceed to traverse further down the tree, if needed.
            # Search queries and access controls apply only at the top level.
            assert not first_level.conditions
            return await first_level.lookup_adapter(segments[1:])

        statement = node_from_segments(segments, root_id=self.node.id)
        statement = self.apply_conditions(statement)  # Conditions on the child node
        statement = statement.options(
            selectinload(orm.Node.data_sources).selectinload(orm.DataSource.structure)
        )

        async with self.context.session() as db:
            node = (await db.execute(statement)).scalar()
        if node is None:
            # Maybe the node does not exist, or maybe we have jumped _inside_ a file
            # whose internal contents are not indexed.

            # TODO As a performance optimization, do binary search or use some
            # tricky JSON query to find where database stops and (for example)
            # HDF5 file begins.

            for i in range(len(segments)):
                catalog_adapter = await self.lookup_adapter(segments[:i])
                if catalog_adapter.data_sources:
                    adapter = await catalog_adapter.get_adapter()
                    for segment in segments[i:]:
                        adapter = await anyio.to_thread.run_sync(adapter.get, segment)
                        if adapter is None:
                            raise NoEntry(segments)
                    return adapter
            raise NoEntry(segments)

        return STRUCTURES[node.structure_family](self.context, node)

    async def get_adapter(self):
        (data_source,) = self.data_sources
        try:
            adapter_cls = self.context.adapters_by_mimetype[data_source.mimetype]
        except KeyError:
            raise RuntimeError(
                f"Server configuration has no adapter for mimetype {data_source.mimetype!r}"
            )
        for asset in data_source.assets:
            if asset.parameter is None:
                continue
            scheme = urlparse(asset.data_uri).scheme
            if scheme == "file":
                # Protect against misbehaving clients reading from unintended parts of the filesystem.
                asset_path = path_from_uri(asset.data_uri)
                for readable_storage in {
                    item
                    for item in self.context.readable_storage
                    if isinstance(item, FileStorage)
                }:
                    if (
                        Path(os.path.commonpath([readable_storage.path, asset_path]))
                        == readable_storage.path
                    ):
                        break
                else:
                    raise RuntimeError(
                        f"Refusing to serve {asset.data_uri} because it is outside "
                        "the readable storage area for this server."
                    )
        adapter = await anyio.to_thread.run_sync(
            partial(
                adapter_cls.from_catalog,
                data_source,
                self.node,
                **data_source.parameters,
            ),
        )
        for query in self.queries:
            if hasattr(adapter, "searc"):
                adapter = adapter.search(query)
        return adapter

    def new_variation(
        self,
        *args,
        sorting=UNCHANGED,
        conditions=UNCHANGED,
        queries=UNCHANGED,
        # must_revalidate=UNCHANGED,
        **kwargs,
    ):
        if sorting is UNCHANGED:
            sorting = self.sorting
        # if must_revalidate is UNCHANGED:
        #     must_revalidate = self.must_revalidate
        if conditions is UNCHANGED:
            conditions = self.conditions
        if queries is UNCHANGED:
            queries = self.queries
        return type(self)(
            self.context,
            node=self.node,
            conditions=conditions,
            sorting=sorting,
            # entries_stale_after=self.entries_stale_after,
            # metadata_stale_after=self.entries_stale_after,
            # must_revalidate=must_revalidate,
            **kwargs,
        )

    def search(self, query):
        if self.data_sources:
            # Stand queries, which will be passed down to the adapter
            # when / if get_adapter() is called.
            self.queries.append(query)
            return self
        return self.query_registry(query, self)

    def sort(self, sorting):
        return self.new_variation(sorting=sorting)

    async def get_distinct(self, metadata, structure_families, specs, counts):
        if self.data_sources:
            return (await self.get_adapter()).get_disinct(
                metadata, structure_families, specs, counts
            )
        data = {}

        async with self.context.session() as db:
            if metadata:
                data["metadata"] = {}
                for key in metadata:
                    clause = orm.Node.metadata_
                    for segment in key.split("."):
                        clause = clause[segment]
                    if counts:
                        columns = (clause, func.count(clause))
                    else:
                        columns = (clause,)
                    statement = select(*columns).group_by(clause)
                    for condition in self.conditions:
                        statement = statement.filter(condition)
                    results = (await db.execute(statement)).all()
                    data["metadata"][key] = format_distinct_result(results, counts)

            if structure_families:
                if counts:
                    columns = (
                        orm.Node.structure_family,
                        func.count(orm.Node.structure_family),
                    )
                else:
                    columns = (orm.Node.structure_family,)
                statement = select(*columns).group_by(orm.Node.structure_family)
                for condition in self.conditions:
                    statement = statement.filter(condition)
                results = (await db.execute(statement)).all()

                data["structure_families"] = format_distinct_result(results, counts)

            if specs:
                if counts:
                    columns = (orm.Node.specs, func.count(orm.Node.specs))
                else:
                    columns = (orm.Node.specs,)
                statement = select(*columns).group_by(orm.Node.specs)
                for condition in self.conditions:
                    statement = statement.filter(condition)
                results = (await db.execute(statement)).all()

                data["specs"] = format_distinct_result(results, counts)

        return data

    @property
    def insert(self):
        # The only way to do "insert if does not exist" i.e. ON CONFLICT
        # is to invoke dialect-specific insert.
        if self.context.engine.dialect.name == "sqlite":
            from sqlalchemy.dialects.sqlite import insert
        elif self.context.engine.dialect.name == "postgresql":
            from sqlalchemy.dialects.postgresql import insert
        else:
            assert False  # future-proofing

        return insert

    async def create_node(
        self,
        structure_family,
        metadata,
        key=None,
        specs=None,
        data_sources=None,
        access_blob=None,
    ):
        access_blob = access_blob or {}
        key = key or self.context.key_maker()
        data_sources = data_sources or []

        node = orm.Node(
            key=key,
            parent=self.node.id,
            metadata_=metadata,
            structure_family=structure_family,
            specs=specs or [],
            access_blob=access_blob,
        )
        async with self.context.session() as db:
            # TODO Consider using nested transitions to ensure that
            # both the node is created (name not already taken)
            # and the directory/file is created---or neither are.
            try:
                db.add(node)
                await db.commit()
            except IntegrityError as exc:
                UNIQUE_CONSTRAINT_FAILED = "gkpj"
                if exc.code == UNIQUE_CONSTRAINT_FAILED:
                    await db.rollback()
                    raise Collision(f"/{'/'.join(await self.path_segments() + [key])}")
                raise
            await db.refresh(node)
            for data_source in data_sources:
                if data_source.management != Management.external:
                    if structure_family == StructureFamily.container:
                        raise NotImplementedError(structure_family)
                    if data_source.mimetype is None:
                        data_source.mimetype = DEFAULT_CREATION_MIMETYPE[
                            data_source.structure_family
                        ]
                    if data_source.mimetype not in STORAGE_ADAPTERS_BY_MIMETYPE:
                        raise HTTPException(
                            status_code=415,
                            detail=(
                                f"The given data source mimetype, {data_source.mimetype}, "
                                "is not one that the Tiled server knows how to write."
                            ),
                        )
                    adapter_cls = STORAGE_ADAPTERS_BY_MIMETYPE[data_source.mimetype]
                    # Choose writable storage. Use the first writable storage item
                    # with a scheme that is supported by this adapter. # For
                    # back-compat, if an adapter does not declare `supported_storage`
                    # assume it supports file-based storage only.
                    supported_storage = getattr(
                        adapter_cls, "supported_storage", {FileStorage}
                    )
                    for storage in self.context.writable_storage:
                        if isinstance(storage, tuple(supported_storage)):
                            break
                    else:
                        raise RuntimeError(
                            f"The adapter {adapter_cls} supports storage types "
                            f"{[cls.__name__ for cls in supported_storage]} "
                            "but the only available storage types "
                            f"are {self.context.writable_storage}."
                        )
                    data_source = await ensure_awaitable(
                        adapter_cls.init_storage,
                        storage,
                        data_source,
                        await self.path_segments() + [key],
                    )
                else:
                    if data_source.mimetype not in self.context.adapters_by_mimetype:
                        raise HTTPException(
                            status_code=HTTP_415_UNSUPPORTED_MEDIA_TYPE,
                            detail=(
                                f"The given data source mimetype, {data_source.mimetype}, "
                                "is not one that the Tiled server knows how to read."
                            ),
                        )

                if data_source.structure is None:
                    structure_id = None
                else:
                    # Obtain and hash the canonical (RFC 8785) representation of
                    # the JSON structure.
                    structure = _prepare_structure(
                        structure_family, data_source.structure
                    )
                    structure_id = compute_structure_id(structure)
                    statement = (
                        self.insert(orm.Structure).values(
                            id=structure_id,
                            structure=structure,
                        )
                    ).on_conflict_do_nothing(index_elements=["id"])
                    await db.execute(statement)
                data_source_orm = orm.DataSource(
                    structure_family=data_source.structure_family,
                    mimetype=data_source.mimetype,
                    management=data_source.management,
                    parameters=data_source.parameters,
                    structure_id=structure_id,
                )
                db.add(data_source_orm)
                node.data_sources.append(data_source_orm)
                await db.flush()  # Get data_source_orm.id.
                for asset in data_source.assets:
                    asset_id = await self._put_asset(db, asset)
                    assoc_orm = orm.DataSourceAssetAssociation(
                        asset_id=asset_id,
                        data_source_id=data_source_orm.id,
                        parameter=asset.parameter,
                        num=asset.num,
                    )
                    db.add(assoc_orm)
            await db.commit()
            # Load with DataSources each DataSource's Structure.
            refreshed_node = (
                await db.execute(
                    select(orm.Node)
                    .filter(orm.Node.id == node.id)
                    .options(
                        selectinload(orm.Node.data_sources).selectinload(
                            orm.DataSource.structure
                        ),
                    )
                )
            ).scalar()
            if self.context.cache_client:
                # Allocate a counter for the new node.
                await self.context.cache_client.setnx(f"sequence:{node.id}", 0)
                # Notify subscribers of the *parent* node about the new child.
                sequence = await self.context.cache_client.incr(
                    f"sequence:{self.node.id}"
                )
                metadata = {
                    "sequence": sequence,
                    "timestamp": datetime.now().isoformat(),
                    "key": key,
                    "structure_family": structure_family,
                    "specs": [spec.model_dump() for spec in (specs or [])],
                    "metadata": metadata,
                    "data_sources": [d.model_dump() for d in data_sources],
                }

                # Cache data in Redis with a TTL, and publish
                # a notification about it.
                pipeline = self.context.cache_client.pipeline()
                pipeline.hset(
                    f"data:{self.node.id}:{sequence}",
                    mapping={
                        "sequence": sequence,
                        "metadata": safe_json_dump(metadata),
                    },
                )
                pipeline.expire(
                    f"data:{self.node.id}:{sequence}", self.context.cache_ttl
                )
                pipeline.publish(f"notify:{self.node.id}", sequence)
                await pipeline.execute()
            return type(self)(self.context, refreshed_node)

    async def _put_asset(self, db: AsyncSession, asset):
        # Find an asset_id if it exists, otherwise create a new one
        statement = select(orm.Asset.id).where(orm.Asset.data_uri == asset.data_uri)
        result = await db.execute(statement)
        if row := result.fetchone():
            (asset_id,) = row
        else:
            statement = self.insert(orm.Asset).values(
                data_uri=asset.data_uri,
                is_directory=asset.is_directory,
            )
            result = await db.execute(statement)
            (asset_id,) = result.inserted_primary_key

        return asset_id

    async def put_data_source(self, data_source, patch):
        # Obtain and hash the canonical (RFC 8785) representation of
        # the JSON structure.
        structure = _prepare_structure(
            data_source.structure_family, data_source.structure
        )
        structure_id = compute_structure_id(structure)
        statement = (
            self.insert(orm.Structure).values(
                id=structure_id,
                structure=structure,
            )
        ).on_conflict_do_nothing(index_elements=["id"])
        async with self.context.session() as db:
            await db.execute(statement)
            values = dict(
                structure_family=data_source.structure_family,
                mimetype=data_source.mimetype,
                management=data_source.management,
                parameters=data_source.parameters,
                structure_id=structure_id,
            )
            result = await db.execute(
                update(orm.DataSource)
                .where(orm.DataSource.id == data_source.id)
                .values(**values)
            )
            if result.rowcount == 0:
                raise HTTPException(
                    status_code=404,
                    detail=f"No data_source {data_source.id} on this node.",
                )
            # Add assets and associate them with the data_source
            for asset in data_source.assets:
                asset_id = await self._put_asset(db, asset)
                statement = select(orm.DataSourceAssetAssociation).where(
                    (orm.DataSourceAssetAssociation.data_source_id == data_source.id)
                    & (orm.DataSourceAssetAssociation.asset_id == asset_id)
                )
                result = await db.execute(statement)
                if not result.fetchone():
                    assoc_orm = orm.DataSourceAssetAssociation(
                        asset_id=asset_id,
                        data_source_id=data_source.id,
                        parameter=asset.parameter,
                        num=asset.num,
                    )
                    db.add(assoc_orm)

            await db.commit()
        if self.context.cache_client:
            sequence = await self.context.cache_client.incr(f"sequence:{self.node.id}")
            metadata = {
                "sequence": sequence,
                "timestamp": datetime.now().isoformat(),
                "data_source": data_source.dict(),
                "patch": patch.dict() if patch else None,
            }

            # Cache data in Redis with a TTL, and publish
            # a notification about it.
            pipeline = self.context.cache_client.pipeline()
            pipeline.hset(
                f"data:{self.node.id}:{sequence}",
                mapping={
                    "sequence": sequence,
                    "metadata": orjson.dumps(metadata),
                },
            )
            pipeline.expire(f"data:{self.node.id}:{sequence}", self.context.cache_ttl)
            pipeline.publish(f"notify:{self.node.id}", sequence)
            await pipeline.execute()

    async def revisions(self, offset, limit):
        async with self.context.session() as db:
            revision_orms = (
                await db.execute(
                    select(orm.Revision)
                    .where(orm.Revision.node_id == self.node.id)
                    .offset(offset)
                    .limit(limit)
                )
            ).all()
            return [Revision.from_orm(o[0]) for o in revision_orms]

    async def delete(self, recursive=False, external_only=True):
        """Delete the Node.

        Any DataSources belonging to this Node and any Assets associated (only) with
        those DataSources will also be deleted.

        If `recursive` is True, delete all Nodes beneath this Node in the tree.
        """
        async with self.context.session() as db:
            if not recursive:
                has_children_stmt = select(
                    exists().where(
                        and_(
                            orm.NodesClosure.ancestor == self.node.id,
                            orm.NodesClosure.descendant != self.node.id,
                        )
                    )
                )
                if (await db.execute(has_children_stmt)).scalar():
                    raise Conflicts(
                        "Cannot delete a node that is not empty. "
                        "Delete its contents first or pass `recursive=True`."
                    )

            affected_nodes_stmnt = (
                select(orm.NodesClosure.descendant)
                .where(orm.NodesClosure.ancestor == self.node.id)
                .distinct()
                .scalar_subquery()
            )
            if external_only:
                int_asset_exists_stmt = select(
                    exists()
                    .where(orm.Asset.id == orm.DataSourceAssetAssociation.asset_id)
                    .where(
                        orm.DataSourceAssetAssociation.data_source_id
                        == orm.DataSource.id
                    )
                    .where(orm.DataSource.node_id.in_(affected_nodes_stmnt))
                    .where(orm.DataSource.management != Management.external)
                )

                if (await db.execute(int_asset_exists_stmt)).scalar():
                    raise WouldDeleteData(
                        "Some items in this tree are internally managed. "
                        "Deleting the records will also delete the underlying data files. "
                        "If you want to delete them, pass external_only=False."
                    )

            sel_asset_stmnt = (
                select(
                    orm.Asset.id,
                    orm.Asset.data_uri,
                    orm.Asset.is_directory,
                    orm.DataSource.management,
                    orm.DataSource.parameters,
                )
                .select_from(orm.Asset)
                .join(
                    orm.Asset.data_sources
                )  # Join on secondary (mapping) relationship
                .join(orm.DataSource.node)
                .filter(orm.Node.id.in_(affected_nodes_stmnt))
                .distinct()
            )

            assets_to_delete = []
            for asset_id, data_uri, is_directory, management, parameters in (
                await db.execute(sel_asset_stmnt)
            ).all():
                # Check if this asset is referenced by other UNAFFECTED nodes
                is_referenced = select(
                    exists()
                    .where(
                        orm.Asset.id == asset_id,
                        orm.Asset.data_sources.any(
                            orm.DataSource.node_id.notin_(affected_nodes_stmnt)
                        ),
                    )
                    .distinct()
                )
                if not (await db.execute(is_referenced)).scalar():
                    # This asset is referenced only by AFFECTED nodes, so we can delete it
                    await db.execute(delete(orm.Asset).where(orm.Asset.id == asset_id))
                    if management != Management.external:
                        assets_to_delete.append((data_uri, is_directory, parameters))
                elif (management == Management.writable) and (
                    urlparse(data_uri).scheme in {"duckdb", "sqlite", "postgresql"}
                ):
                    # The tabular storage asset may be referenced by several data_sources
                    # and nodes, so we cannot delete it completely. However, we can delete
                    # the relevant rows and tables.
                    assets_to_delete.append((data_uri, is_directory, parameters))

            result = await db.execute(
                delete(orm.Node)
                .where(orm.Node.id.in_(affected_nodes_stmnt))
                .where(orm.Node.parent.isnot(None))
            )
            await db.commit()

            # Finally, delete the physical assets that are not externally managed
            for data_uri, is_directory, parameters in assets_to_delete:
                delete_asset(data_uri, is_directory, parameters=parameters)

        return result.rowcount

    async def delete_revision(self, number):
        async with self.context.session() as db:
            result = await db.execute(
                delete(orm.Revision)
                .where(orm.Revision.node_id == self.node.id)
                .where(orm.Revision.revision_number == number)
            )
            if result.rowcount == 0:
                # TODO Abstract this from FastAPI?
                raise HTTPException(
                    status_code=HTTP_404_NOT_FOUND,
                    detail=f"No revision {number} for node {self.node.id}",
                )
            assert (
                result.rowcount == 1
            ), f"Deletion would affect {result.rowcount} rows; rolling back"
            await db.commit()

    async def replace_metadata(
        self, metadata=None, specs=None, access_blob=None, *, drop_revision=False
    ):
        values = {}
        if metadata is not None:
            # Trailing underscore in 'metadata_' avoids collision with
            # SQLAlchemy reserved word 'metadata'.
            values["metadata_"] = metadata
        if specs is not None:
            values["specs"] = specs
        if access_blob is not None:
            values["access_blob"] = access_blob
        async with self.context.session() as db:
            if not drop_revision:
                current = (
                    await db.execute(
                        select(orm.Node).where(orm.Node.id == self.node.id)
                    )
                ).scalar_one()
                next_revision_number = 1 + (
                    (
                        await db.execute(
                            select(func.max(orm.Revision.revision_number)).where(
                                orm.Revision.node_id == self.node.id
                            )
                        )
                    ).scalar_one()
                    or 0
                )
                revision = orm.Revision(
                    # Trailing underscore in 'metadata_' avoids collision with
                    # SQLAlchemy reserved word 'metadata'.
                    metadata_=current.metadata_,
                    specs=current.specs,
                    access_blob=current.access_blob,
                    node_id=current.id,
                    revision_number=next_revision_number,
                )
                db.add(revision)
            await db.execute(
                update(orm.Node).where(orm.Node.id == self.node.id).values(**values)
            )
            await db.commit()

    async def close_stream(self):
        # Check the node status.
        # ttl returns -2 if the key does not exist.
        # ttl returns -1 if the key exists but has no associated expire.
        # ttl greater than 0 means that it is marked to expire.
        node_ttl = await self.context.cache_client.ttl(f"sequence:{self.node.id}")
        if node_ttl > 0:
            raise HTTPException(
                status_code=HTTP_404_NOT_FOUND,
                detail=f"Stream for node {self.node.id} is already closed.",
            )
        if node_ttl == -2:
            raise HTTPException(
                status_code=HTTP_404_NOT_FOUND,
                detail=f"Node {self.node.id} not found.",
            )

        metadata = {
            "timestamp": datetime.now().isoformat(),
            "end_of_stream": True,
        }
        # Increment the counter for this node.
        sequence = await self.context.cache_client.incr(f"sequence:{self.node.id}")

        # Cache data in Redis with a TTL, and publish
        # a notification about it.
        pipeline = self.context.cache_client.pipeline()
        pipeline.hset(
            f"data:{self.node.id}:{sequence}",
            mapping={
                "sequence": sequence,
                "metadata": orjson.dumps(metadata),
            },
        )
        pipeline.expire(f"data:{self.node.id}:{sequence}", self.context.cache_ttl)
        pipeline.expire(f"sequence:{self.node.id}", self.context.cache_ttl)
        pipeline.publish(f"notify:{self.node.id}", sequence)
        await pipeline.execute()

    def make_ws_handler(self, websocket, formatter, uri):
        async def handler(sequence: Optional[int] = None):
            await websocket.accept()
            end_stream = asyncio.Event()
            cache_client = self.context.cache_client

            async def stream_data(sequence):
                key = f"data:{self.node.id}:{sequence}"
                payload_bytes, metadata_bytes = await cache_client.hmget(
                    key, "payload", "metadata"
                )
                if metadata_bytes is None:
                    # This means that redis ttl has expired for this sequence
                    return
                metadata = orjson.loads(metadata_bytes)
                if metadata.get("end_of_stream"):
                    # This means that the stream is closed by the producer
                    end_stream.set()
                    return
                metadata["uri"] = uri
                if metadata.get("patch"):
                    s = ",".join(
                        f"{offset}:{offset+shape}"
                        for offset, shape in zip(
                            metadata["patch"]["offset"], metadata["patch"]["shape"]
                        )
                    )
                    metadata["uri"] = f"{uri}?slice={s}"
                await formatter(websocket, metadata, payload_bytes)

            # Setup buffer
            stream_buffer = asyncio.Queue()

            async def buffer_live_events():
                pubsub = cache_client.pubsub()
                await pubsub.subscribe(f"notify:{self.node.id}")
                try:
                    async for message in pubsub.listen():
                        if message.get("type") == "message":
                            try:
                                live_seq = int(message["data"])
                                await stream_buffer.put(live_seq)
                            except Exception as e:
                                print(f"Error parsing live message: {e}")
                except Exception as e:
                    print(f"Live subscription error: {e}")
                finally:
                    await pubsub.unsubscribe(f"notify:{self.node.id}")
                    await pubsub.aclose()

            live_task = asyncio.create_task(buffer_live_events())

            if sequence is not None:
                current_seq = await cache_client.get(f"sequence:{self.node.id}")
                current_seq = int(current_seq) if current_seq is not None else 0
                print("Replaying old data...")
                for s in range(sequence, current_seq + 1):
                    await stream_data(s)
            # New data
            try:
                while not end_stream.is_set():
                    live_seq = await stream_buffer.get()
                    await stream_data(live_seq)
                else:
                    await websocket.close(code=1000, reason="Producer ended stream")
            except WebSocketDisconnect:
                print(f"Client disconnected from node {self.node.id}")
            finally:
                live_task.cancel()

        return handler


class CatalogContainerAdapter(CatalogNodeAdapter):
    async def keys_range(self, offset, limit):
        if self.data_sources:
            return it.islice(
                (await self.get_adapter()).keys(),
                offset,
                (offset + limit) if limit is not None else None,  # noqa: E203
            )
        statement = select(orm.Node.key).filter(orm.Node.parent == self.node.id)
        statement = self.apply_conditions(statement)
        async with self.context.session() as db:
            return (
                (
                    await db.execute(
                        statement.order_by(*self.order_by_clauses)
                        .offset(offset)
                        .limit(limit)
                    )
                )
                .scalars()
                .all()
            )

    async def items_range(self, offset, limit):
        if self.data_sources:
            return it.islice(
                (await self.get_adapter()).items(),
                offset,
                (offset + limit) if limit is not None else None,  # noqa: E203
            )
        statement = select(orm.Node).filter(orm.Node.parent == self.node.id)
        statement = self.apply_conditions(statement)
        async with self.context.session() as db:
            nodes = (
                (
                    await db.execute(
                        statement.order_by(*self.order_by_clauses)
                        .offset(offset)
                        .limit(limit)
                    )
                )
                .scalars()
                .all()
            )
            return [
                (
                    node.key,
                    STRUCTURES[node.structure_family](self.context, node),
                )
                for node in nodes
            ]

    async def read(self, *args, **kwargs):
        if not self.data_sources:
            fields = kwargs.get("fields")
            if fields:
                return self.search(KeysFilter(fields))
            return self
        return await ensure_awaitable((await self.get_adapter()).read, *args, **kwargs)


class CatalogArrayAdapter(CatalogNodeAdapter):
    async def read(self, *args, **kwargs):
        if not self.data_sources:
            fields = kwargs.get("fields")
            if fields:
                return self.search(KeysFilter(fields))
            return self
        return await ensure_awaitable((await self.get_adapter()).read, *args, **kwargs)

    async def read_block(self, *args, **kwargs):
        return await ensure_awaitable(
            (await self.get_adapter()).read_block, *args, **kwargs
        )

    async def _stream(self, media_type, entry, body, shape, block=None, offset=None):
        sequence = await self.context.cache_client.incr(f"sequence:{self.node.id}")
        metadata = {
            "timestamp": datetime.now().isoformat(),
            "content-type": media_type,
            "shape": shape,
            "offset": offset,
            "block": block,
        }

        pipeline = self.context.cache_client.pipeline()
        pipeline.hset(
            f"data:{self.node.id}:{sequence}",
            mapping={
                "sequence": sequence,
                "metadata": orjson.dumps(metadata),
                "payload": body,  # raw user input
            },
        )
        pipeline.expire(f"data:{self.node.id}:{sequence}", self.context.cache_ttl)
        pipeline.publish(f"notify:{self.node.id}", sequence)
        await pipeline.execute()

    async def write(self, media_type, deserializer, entry, body):
        shape = entry.structure().shape
        if self.context.cache_client:
            await self._stream(media_type, entry, body, shape)
        if entry.structure_family == "array":
            dtype = entry.structure().data_type.to_numpy_dtype()
            data = await ensure_awaitable(deserializer, body, dtype, shape)
        elif entry.structure_family == "sparse":
            data = await ensure_awaitable(deserializer, body)
        else:
            raise NotImplementedError(entry.structure_family)
        return await ensure_awaitable((await self.get_adapter()).write, data)

    async def write_block(self, block, media_type, deserializer, entry, body):
        from tiled.adapters.array import slice_and_shape_from_block_and_chunks

        _, shape = slice_and_shape_from_block_and_chunks(
            block, entry.structure().chunks
        )
        if self.context.cache_client:
            await self._stream(media_type, entry, body, shape, block=block)
        if entry.structure_family == "array":
            dtype = entry.structure().data_type.to_numpy_dtype()
            data = await ensure_awaitable(deserializer, body, dtype, shape)
        elif entry.structure_family == "sparse":
            data = await ensure_awaitable(deserializer, body)
        else:
            raise NotImplementedError(entry.structure_family)
        return await ensure_awaitable(
            (await self.get_adapter()).write_block, data, block
        )

    async def patch(self, shape, offset, extend, media_type, deserializer, entry, body):
        if self.context.cache_client:
            await self._stream(media_type, entry, body, shape, offset=offset)
        dtype = entry.structure().data_type.to_numpy_dtype()
        data = await ensure_awaitable(deserializer, body, dtype, shape)
        # assumes a single DataSource (currently only supporting zarr)
        async with self.context.session() as db:
            new_shape_and_chunks = await ensure_awaitable(
                (await self.get_adapter()).patch, data, offset, extend
            )
            node = await db.get(orm.Node, self.node.id)
            if len(node.data_sources) != 1:
                raise NotImplementedError("Only handles one data source")
            data_source = node.data_sources[0]
            structure_row = await db.get(orm.Structure, data_source.structure_id)
            # Get the current structure row, update the shape, and write it back
            structure_dict = copy.deepcopy(structure_row.structure)
            structure_dict["shape"], structure_dict["chunks"] = new_shape_and_chunks
            new_structure_id = compute_structure_id(structure_dict)
            statement = (
                self.insert(orm.Structure)
                .values(
                    id=new_structure_id,
                    structure=structure_dict,
                )
                .on_conflict_do_nothing(index_elements=["id"])
            )
            await db.execute(statement)
            new_structure = await db.get(orm.Structure, new_structure_id)
            data_source.structure = new_structure
            data_source.structure_id = new_structure_id
            db.add(data_source)
            await db.commit()
            return structure_dict


class CatalogAwkwardAdapter(CatalogNodeAdapter):
    async def read(self, *args, **kwargs):
        return await ensure_awaitable((await self.get_adapter()).read, *args, **kwargs)

    async def read_buffers(self, *args, **kwargs):
        return await ensure_awaitable(
            (await self.get_adapter()).read_buffers, *args, **kwargs
        )

    async def write(self, *args, **kwargs):
        return await ensure_awaitable((await self.get_adapter()).write, *args, **kwargs)


class CatalogSparseAdapter(CatalogArrayAdapter):
    pass


class CatalogTableAdapter(CatalogNodeAdapter):
    async def get(self, *args, **kwargs):
        return (await self.get_adapter()).get(*args, **kwargs)

    async def read(self, *args, **kwargs):
        return await ensure_awaitable((await self.get_adapter()).read, *args, **kwargs)

    async def write(self, *args, **kwargs):
        return await ensure_awaitable((await self.get_adapter()).write, *args, **kwargs)

    async def read_partition(self, *args, **kwargs):
        return await ensure_awaitable(
            (await self.get_adapter()).read_partition, *args, **kwargs
        )

    async def write_partition(self, *args, **kwargs):
        return await ensure_awaitable(
            (await self.get_adapter()).write_partition, *args, **kwargs
        )

    async def append_partition(self, *args, **kwargs):
        return await ensure_awaitable(
            (await self.get_adapter()).append_partition, *args, **kwargs
        )


def delete_asset(data_uri, is_directory, parameters=None):
    url = urlparse(data_uri)
    if url.scheme == "file":
        path = path_from_uri(data_uri)
        if is_directory:
            shutil.rmtree(path)
        else:
            Path(path).unlink()
    elif url.scheme in {"duckdb", "sqlite", "postgresql"}:
        storage = cast(SQLStorage, get_storage(data_uri))
        with closing(storage.connect()) as conn:
            table_name = parameters.get("table_name") if parameters else None
            dataset_id = parameters.get("dataset_id") if parameters else None
            with conn.cursor() as cursor:
                cursor.execute(
                    f'DELETE FROM "{table_name}" WHERE _dataset_id = {dataset_id:d};',
                )
            conn.commit()

            # If the table is empty, we can drop it
            with conn.cursor() as cursor:
                cursor.execute(f'SELECT COUNT(*) FROM "{table_name}";')
                if cursor.fetchone()[0] == 0:
                    cursor.execute(f'DROP TABLE IF EXISTS "{table_name}";')
            conn.commit()
    else:
        raise NotImplementedError(
            f"Cannot delete asset at {data_uri!r} because the scheme {url.scheme!r} is not supported."
        )


_STANDARD_SORT_KEYS = {
    "id": "key",
    # Maybe add things like structure_family here, but it's not clear what the
    # sort order would be.
}


def order_by_clauses(sorting):
    clauses = []
    default_sorting_direction = 1
    for key, direction in sorting:
        if key == "":
            default_sorting_direction = direction
            continue
            # TODO Really we should insist that if this is given, it is last,
            # because we always apply the default sorting last.
        if key in _STANDARD_SORT_KEYS:
            clause = getattr(orm.Node, _STANDARD_SORT_KEYS[key])
        else:
            clause = orm.Node.metadata_
            # This can be given bare like "color" or namedspaced like
            # "metadata.color" to avoid the possibility of a name collision
            # with the standard sort keys.
            if key.startswith("metadata."):
                key = key[len("metadata.") :]  # noqa: E203

            for segment in key.split("."):
                clause = clause[segment]
        if direction == -1:
            clause = clause.desc()
        clauses.append(clause)
    # Ensure deterministic ordering for all queries by sorting by
    # 'time_created' and then by 'id' last.
    for clause in [orm.Node.time_created, orm.Node.id]:
        if default_sorting_direction == -1:
            clause = clause.desc()
        clauses.append(clause)
    return clauses


_TYPE_CONVERSION_MAP = {
    int: "as_integer",
    float: "as_float",
    str: "as_string",
    bool: "as_boolean",
}


def _get_value(value, type):
    # This is used only for SQLite, to get the types right
    # so that that index is used. There is probably a
    # cleaner way to handle this.
    # Study https://gist.github.com/brthor/e3d23ae549ee53cdea56d72d39ad1288
    # which may or may not be relevant anymore.
    return getattr(value, _TYPE_CONVERSION_MAP[type])()


def _prepare_structure(structure_family, structure):
    "Convert from pydantic model to dict."
    if structure is None:
        return None
    if isinstance(structure, dict):
        return structure
    return dataclasses.asdict(structure)


def binary_op(query, tree, operation):
    dialect_name = tree.context.engine.url.get_dialect().name
    keys = query.key.split(".")
    attr = orm.Node.metadata_[keys]
    if dialect_name == "sqlite":
        condition = operation(_get_value(attr, type(query.value)), query.value)
    # specific case where GIN optomized index can be used to speed up POSTGRES equals queries
    elif (dialect_name == "postgresql") and (operation == operator.eq):
        condition = orm.Node.metadata_.op("@>")(
            type_coerce(
                key_array_to_json(keys, query.value),
                orm.Node.metadata_.type,
            )
        )
    else:
        condition = operation(attr, type_coerce(query.value, orm.Node.metadata_.type))
    return tree.new_variation(conditions=tree.conditions + [condition])


def like(query, tree):
    keys = query.key.split(".")
    attr = orm.Node.metadata_[keys]
    condition = _get_value(attr, str).like(query.pattern)
    return tree.new_variation(conditions=tree.conditions + [condition])


def comparison(query, tree):
    OPERATORS = {
        Operator.lt: operator.lt,
        Operator.le: operator.le,
        Operator.gt: operator.gt,
        Operator.ge: operator.ge,
    }
    return binary_op(query, tree, OPERATORS[query.operator])


def contains(query, tree):
    attr = orm.Node.metadata_[query.key.split(".")]
    condition = _get_value(attr, type(query.value)).contains(query.value)
    return tree.new_variation(conditions=tree.conditions + [condition])


def full_text(query, tree):
    dialect_name = tree.context.engine.url.get_dialect().name
    if dialect_name == "sqlite":
        condition = orm.metadata_fts5.c.metadata.match(query.text)
    elif dialect_name == "postgresql":
        tsvector = func.jsonb_to_tsvector(
            sql_cast("simple", REGCONFIG),
            orm.Node.metadata_,
            sql_cast(["string"], JSONB),
        )
        condition = tsvector.op("@@")(func.to_tsquery("simple", query.text))
    else:
        raise UnsupportedQueryType("full_text")
    return tree.new_variation(conditions=tree.conditions + [condition])


def specs(query, tree):
    dialect_name = tree.context.engine.url.get_dialect().name
    conditions = []
    attr = orm.Node.specs
    if dialect_name == "sqlite":
        # Construct the conditions for includes
        for i, name in enumerate(query.include):
            conditions.append(attr.like(f'%{{"name":"{name}",%'))
        # Construct the conditions for excludes
        for i, name in enumerate(query.exclude):
            conditions.append(not_(attr.like(f'%{{"name":"{name}",%')))
    elif dialect_name == "postgresql":
        if query.include:
            conditions.append(attr.op("@>")(specs_array_to_json(query.include)))
        if query.exclude:
            conditions.append(not_(attr.op("@>")(specs_array_to_json(query.exclude))))
    else:
        raise UnsupportedQueryType("specs")
    return tree.new_variation(conditions=tree.conditions + conditions)


def access_blob_filter(query, tree):
    dialect_name = tree.context.engine.url.get_dialect().name
    access_blob = orm.Node.access_blob
    if not (query.user_id or query.tags):
        # Results cannot possibly match an empty value or list,
        # so put a False condition in the list ensuring that
        # there are no rows returned.
        condition = false()
    elif dialect_name == "sqlite":
        attr_id = access_blob["user"]
        attr_tags = access_blob["tags"]
        access_tags_json = func.json_each(attr_tags).table_valued("value")
        condition = (
            select(1)
            .select_from(access_tags_json)
            .where(access_tags_json.c.value.in_(query.tags))
            .exists()
        )
        if query.user_id is not None:
            user_match = (
                func.json_extract(func.json_quote(attr_id), "$") == query.user_id
            )
            condition = or_(condition, user_match)
    elif dialect_name == "postgresql":
        access_blob_jsonb = type_coerce(access_blob, JSONB)
        condition = access_blob_jsonb["tags"].has_any(sql_cast(query.tags, ARRAY(TEXT)))
        if query.user_id is not None:
            user_match = access_blob_jsonb["user"].astext == query.user_id
            condition = or_(condition, user_match)
    else:
        raise UnsupportedQueryType("access_blob_filter")

    return tree.new_variation(conditions=tree.conditions + [condition])


def in_or_not_in_sqlite(query, tree, method):
    keys = query.key.split(".")
    attr = orm.Node.metadata_[keys]
    if len(query.value) == 0:
        if method == "in_":
            # Results cannot possibly be "in" in an empty list,
            # so put a False condition in the list ensuring that
            # there are no rows return.
            condition = false()
        else:  # method == "not_in"
            # All results are always "not in" an empty list.
            condition = true()
    else:
        condition = getattr(_get_value(attr, type(query.value[0])), method)(query.value)
    return tree.new_variation(conditions=tree.conditions + [condition])


def in_or_not_in_postgresql(query, tree, method):
    keys = query.key.split(".")
    # Engage btree_gin index with @> operator
    if method == "in_":
        if len(query.value) == 0:
            condition = false()
        else:
            condition = or_(
                *(
                    orm.Node.metadata_.op("@>")(key_array_to_json(keys, item))
                    for item in query.value
                )
            )
    elif method == "not_in":
        if len(query.value) == 0:
            condition = true()
        else:
            condition = not_(
                or_(
                    *(
                        orm.Node.metadata_.op("@>")(key_array_to_json(keys, item))
                        for item in query.value
                    )
                )
            )
    return tree.new_variation(conditions=tree.conditions + [condition])


_IN_OR_NOT_IN_DIALECT_DISPATCH: Dict[str, Callable] = {
    "sqlite": in_or_not_in_sqlite,
    "postgresql": in_or_not_in_postgresql,
}


def in_or_not_in(query, tree, method):
    METHODS = {"in_", "not_in"}
    if method not in METHODS:
        raise ValueError(f"method must be one of {METHODS}")
    dialect_name = tree.context.engine.url.get_dialect().name
    return _IN_OR_NOT_IN_DIALECT_DISPATCH[dialect_name](query, tree, method)


def key_present(query, tree):
    # Functionally in SQLAlchemy 'is not None' does not work as expected
    if tree.context.engine.url.get_dialect().name == "sqlite":
        condition = orm.Node.metadata_.op("->")("$." + query.key) != None  # noqa: E711
    else:
        keys = query.key.split(".")
        condition = (
            orm.Node.metadata_.op("#>")(sql_cast(keys, ARRAY(TEXT)))
            != None  # noqa: E711
        )
    condition = condition if getattr(query, "exists", True) else not_(condition)
    return tree.new_variation(conditions=tree.conditions + [condition])


def keys_filter(query, tree):
    condition = orm.Node.key.in_(query.keys)
    return tree.new_variation(conditions=tree.conditions + [condition])


def structure_family(query, tree):
    condition = orm.Node.structure_family == query.value
    return tree.new_variation(conditions=tree.conditions + [condition])


CatalogNodeAdapter.register_query(Eq, partial(binary_op, operation=operator.eq))
CatalogNodeAdapter.register_query(NotEq, partial(binary_op, operation=operator.ne))
CatalogNodeAdapter.register_query(Comparison, comparison)
CatalogNodeAdapter.register_query(Contains, contains)
CatalogNodeAdapter.register_query(In, partial(in_or_not_in, method="in_"))
CatalogNodeAdapter.register_query(NotIn, partial(in_or_not_in, method="not_in"))
CatalogNodeAdapter.register_query(KeyPresent, key_present)
CatalogNodeAdapter.register_query(KeysFilter, keys_filter)
CatalogNodeAdapter.register_query(StructureFamilyQuery, structure_family)
CatalogNodeAdapter.register_query(SpecsQuery, specs)
CatalogNodeAdapter.register_query(AccessBlobFilter, access_blob_filter)
CatalogNodeAdapter.register_query(FullText, full_text)
CatalogNodeAdapter.register_query(Like, like)


def in_memory(
    *,
    named_memory=None,
    metadata=None,
    specs=None,
    writable_storage=None,
    readable_storage=None,
    echo=DEFAULT_ECHO,
    adapters_by_mimetype=None,
    top_level_access_blob=None,
    cache_settings=None,
):
    if not named_memory:
        uri = "sqlite:///:memory:"
    else:
        uri = f"sqlite:///file:{named_memory}?mode=memory&cache=shared&uri=true"
    # NOTE: catalog_pool_size and catalog_max_overflow are ignored when using an
    # in-memory catalog.
    return from_uri(
        uri=uri,
        metadata=metadata,
        specs=specs,
        writable_storage=writable_storage,
        readable_storage=readable_storage,
        init_if_not_exists=True,
        echo=echo,
        adapters_by_mimetype=adapters_by_mimetype,
        top_level_access_blob=top_level_access_blob,
        cache_settings=cache_settings,
    )


def from_uri(
    uri,
    *,
    metadata=None,
    specs=None,
    writable_storage=None,
    readable_storage=None,
    init_if_not_exists=False,
    echo=DEFAULT_ECHO,
    adapters_by_mimetype=None,
    top_level_access_blob=None,
    mount_node: Optional[Union[str, List[str]]] = None,
    cache_settings=None,
    catalog_pool_size=None,
    storage_pool_size=None,
    catalog_max_overflow=None,
    storage_max_overflow=None,
):
    uri = ensure_specified_sql_driver(uri)
    if init_if_not_exists:
        # The alembic stamping can only be does synchronously.
        # The cleanest option available is to start a subprocess
        # because SQLite is allergic to threads.
        import subprocess

        # TODO Check if catalog exists.
        process = subprocess.run(
            [sys.executable, "-m", "tiled", "catalog", "init", "--if-not-exists", uri],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
        )
        # Capture stdout and stderr from the subprocess and write to logging
        stdout = process.stdout.decode()
        stderr = process.stderr.decode()
        logger.info(f"Subprocess stdout: {stdout}")
        logger.info(f"Subprocess stderr: {stderr}")

    parsed_url = make_url(uri)
    if (
        (parsed_url.get_dialect().name == "sqlite")
        and (parsed_url.database != ":memory:")
        and (parsed_url.query.get("mode", None) != "memory")
    ):
        # For file-backed SQLite databases, connection pooling offers a
        # significant performance boost. For SQLite databases that exist
        # only in process memory, pooling is not applicable.
        poolclass = AsyncAdaptedQueuePool
    elif parsed_url.get_dialect().name.startswith("postgresql"):
        poolclass = AsyncAdaptedQueuePool  # Default for PostgreSQL
    else:
        poolclass = None  # defer to sqlalchemy default

    node = RootNode(metadata, specs, top_level_access_blob)
    mount_path = (
        [segment for segment in mount_node.split("/") if segment]
        if isinstance(mount_node, str)
        else mount_node
    )

    # Optionally set pool size and max overflow for the catalog engine;
    # if not specified, use the default values from sqlalchemy.
    pool_kwargs = (
        {"pool_size": catalog_pool_size, "max_overflow": catalog_max_overflow}
        if poolclass == AsyncAdaptedQueuePool
        else {}
    )
    pool_kwargs = {k: v for k, v in pool_kwargs.items() if v is not None}
    # Create the async engine with the specified parameters
    engine = create_async_engine(
        uri,
        echo=echo,
        json_serializer=json_serializer,
        poolclass=poolclass,
        **pool_kwargs,
    )
    if engine.dialect.name == "sqlite":
        event.listens_for(engine.sync_engine, "connect")(_set_sqlite_pragma)
    context = Context(
        engine,
        writable_storage,
        readable_storage,
        adapters_by_mimetype,
        cache_settings,
        storage_pool_size=storage_pool_size,
        storage_max_overflow=storage_max_overflow,
    )
    adapter = CatalogContainerAdapter(context, node, mount_path=mount_path)

    return adapter


def _set_sqlite_pragma(conn, record):
    cursor = conn.cursor()
    # https://docs.sqlalchemy.org/en/13/dialects/sqlite.html#foreign-key-support
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


def format_distinct_result(results, counts):
    if counts:
        formatted_result = [
            {"value": value, "count": count} for value, count in results
        ]
    else:
        formatted_result = [{"value": value} for value, in results]
    return formatted_result


class WouldDeleteData(RuntimeError):
    pass


class Collision(Conflicts):
    pass


def json_serializer(obj):
    "The PostgreSQL JSON serializer requires str, not bytes."
    return safe_json_dump(obj).decode()


def key_array_to_json(keys, value):
    """Take JSON accessor information as an array of keys and value

    Parameters
    ----------
    keys : iterable
        An array of keys to be created in the object.
    value : string
        Value assigned to the final key.

    Returns
    -------
    json
        JSON object for use in postgresql queries.

    Examples
    --------
    >>> key_array_to_json(['x','y','z'], 1)
    {'x': {'y': {'z': 1}}
    """
    return {keys[0]: reduce(lambda x, y: {y: x}, keys[1:][::-1], value)}


def specs_array_to_json(specs):
    """Take array of Specs strings and convert them to a `penguin` @> friendly array
    Assume constructed array will feature keys called "name"

    Parameters
    ----------
    specs : iterable
        An array of specs strings to be searched for.

    Returns
    -------
    json
        JSON object for use in postgresql queries.

    Examples
    --------
    >>> specs_array_to_json(['foo','bar'])
    [{"name":"foo"},{"name":"bar"}]
    """
    return [{"name": spec} for spec in specs]


def node_from_segments(segments, root_id=0):
    """Create an SQLAlchemy select statement to find a node based on its path

    Queries the database recursively to find the node with the given ancestors
    and key.

    Parameters
    ----------
        segments : list of str
            The path segments leading to the node, e.g. ['A', 'x', 'i'].
        root_id : int
            The ID of the root node, typically 0 for the root of the catalog.

    Returns
    -------
        sqlalchemy.sql.selectable.Select
    """

    # Create an alias for each ancestor node in the path and build the join chain
    orm_NodeAliases = [aliased(orm.Node) for _ in range(len(segments))] + [orm.Node]
    statement = select(orm_NodeAliases[-1])  # Select the child node
    statement = statement.select_from(orm_NodeAliases[0])  # Start from the ancestor
    statement = statement.where(orm_NodeAliases[0].id == root_id)
    for i, segment in enumerate(segments):
        parent, child = orm_NodeAliases[i], orm_NodeAliases[i + 1]
        statement = statement.join(child, child.parent == parent.id).where(
            child.key == segment
        )

    return statement


STRUCTURES = {
    StructureFamily.array: CatalogArrayAdapter,
    StructureFamily.awkward: CatalogAwkwardAdapter,
    StructureFamily.container: CatalogContainerAdapter,
    StructureFamily.sparse: CatalogSparseAdapter,
    StructureFamily.table: CatalogTableAdapter,
}



import json
import sqlite3
from copy import deepcopy

import numpy
import pytest
from starlette.status import HTTP_403_FORBIDDEN

from ..access_control.access_tags import AccessTagsCompiler
from ..access_control.scopes import ALL_SCOPES
from ..client import Context, from_context
from ..server.app import build_app_from_config
from .utils import enter_username_password, fail_with_status_code

arr = numpy.ones((5, 5))


server_config = {
    "access_control": {
        "access_policy": "tiled.access_control.access_policies:TagBasedAccessPolicy",
        "args": {
            "provider": "toy",
            "tags_db": {
                "uri": "file:compiled_tags_mem?mode=memory&cache=shared"  # in-memory and shareable
            },
            "access_tags_parser": "tiled.access_control.access_tags:AccessTagsParser",
        },
    },
    "authentication": {
        "tiled_admins": [{"provider": "toy", "id": "admin"}],
        "allow_anonymous_access": True,
        "secret_keys": ["SECRET"],
        "providers": [
            {
                "provider": "toy",
                "authenticator": "tiled.authenticators:DictionaryAuthenticator",
                "args": {
                    "users_to_passwords": {
                        "alice": "alice",
                        "bob": "bob",
                        "chris": "chris",
                        "sue": "sue",
                        "zoe": "zoe",
                        "admin": "admin",
                    },
                },
            },
        ],
    },
    "database": {
        "uri": "sqlite:///file:authn_mem?mode=memory&cache=shared&uri=true",  # in-memory
    },
}

access_tag_config = {
    "roles": {
        "facility_user": {
            "scopes": [
                "read:data",
                "read:metadata",
            ]
        },
        "facility_admin": {
            "scopes": [
                "read:data",
                "read:metadata",
                "write:data",
                "write:metadata",
                "create",
                "register",
            ]
        },
    },
    "tags": {
        "alice_tag": {
            "users": [
                {
                    "name": "alice",
                    "role": "facility_admin",
                },
                {
                    "name": "chris",
                    "scopes": ["read:data", "read:metadata"],
                },
            ],
        },
        "chris_tag": {
            "users": [
                {
                    "name": "alice",
                    "role": "facility_admin",
                },
                {
                    "name": "chris",
                    "role": "facility_admin",
                },
            ],
        },
        "biologists_tag": {
            "users": [
                {
                    "name": "alice",
                    "role": "facility_admin",
                },
            ],
            "groups": [
                {
                    "name": "biologists",
                    "scopes": ["read:data", "read:metadata"],
                },
            ],
        },
        "chemists_tag": {
            "users": [
                {
                    "name": "sue",
                    "scopes": ["write:data", "write:metadata"],
                },
            ],
            "groups": [
                {
                    "name": "chemists",
                    "role": "facility_user",
                },
            ],
            "auto_tags": [
                {
                    "name": "alice_tag",
                },
            ],
        },
        "physicists_tag": {
            "users": [
                {
                    "name": "alice",
                    "role": "facility_admin",
                },
            ],
            "groups": [
                {
                    "name": "physicists",
                    "role": "facility_admin",
                },
            ],
        },
    },
    "tag_owners": {
        "alice_tag": {
            "users": [
                {
                    "name": "alice",
                },
                {
                    "name": "chris",
                },
            ],
        },
        "biologists_tag": {
            "users": [
                {
                    "name": "alice",
                },
            ],
            "groups": [
                {
                    "name": "biologists",
                },
            ],
        },
        "chemists_tag": {
            "users": [
                {
                    "name": "sue",
                },
            ],
            "groups": [
                {
                    "name": "chemists",
                },
            ],
        },
        "physicists_tag": {
            "users": [
                {
                    "name": "alice",
                },
            ],
            "groups": [
                {
                    "name": "physicists",
                },
            ],
        },
    },
}


def group_parser(groupname):
    return {
        "chemists": ["bob", "mary"],
        "biologists": ["chris", "fred"],
        "physicists": ["sue", "tony"],
    }[groupname]


@pytest.fixture(scope="module")
def compile_access_tags_db():
    access_tags_compiler = AccessTagsCompiler(
        ALL_SCOPES,
        access_tag_config,
        {"uri": "file:compiled_tags_mem?mode=memory&cache=shared"},
        group_parser,
    )

    access_tags_compiler.load_tag_config()
    access_tags_compiler.compile()
    yield access_tags_compiler
    access_tags_compiler.connection.close()


@pytest.fixture
def compile_access_tags_db_with_reset(compile_access_tags_db):
    access_tags_compiler = compile_access_tags_db
    access_tag_config_copy = deepcopy(access_tag_config)
    access_tags_compiler.tag_config = access_tag_config_copy
    yield access_tags_compiler
    access_tags_compiler.tag_config = access_tag_config
    access_tags_compiler.group_parser = group_parser
    access_tags_compiler.clear_raw_tags()
    access_tags_compiler.load_tag_config()
    access_tags_compiler.recompile()


@pytest.fixture(scope="module")
def access_control_test_context_factory(tmpdir_module, compile_access_tags_db):
    config = {
        "trees": [
            {
                "tree": "tiled.catalog:in_memory",
                "args": {
                    "named_memory": "catalog_foo",
                    "writable_storage": str(tmpdir_module / "foo"),
                    "top_level_access_blob": {"tags": ["alice_tag"]},
                },
                "path": "/foo",
            },
            {
                "tree": "tiled.catalog:in_memory",
                "args": {
                    "named_memory": "catalog_bar",
                    "writable_storage": str(tmpdir_module / "bar"),
                    "top_level_access_blob": {"tags": ["chemists_tag"]},
                },
                "path": "/bar",
            },
            {
                "tree": "tiled.catalog:in_memory",
                "args": {
                    "named_memory": "catalog_baz",
                    "writable_storage": str(tmpdir_module / "baz"),
                    "top_level_access_blob": {"tags": ["physicists_tag"]},
                },
                "path": "/baz",
            },
            {
                "tree": "tiled.catalog:in_memory",
                "args": {
                    "named_memory": "catalog_qux",
                    "writable_storage": str(tmpdir_module / "qux"),
                    "top_level_access_blob": {"tags": ["public"]},
                },
                "path": "/qux",
            },
        ],
    }

    config.update(server_config)
    contexts = []
    clients = {}

    def _create_and_login_context(username, password=None, api_key=None):
        if not any([password, api_key]):
            raise ValueError("Please provide either 'password' or 'api_key' for auth")

        if client := clients.get(username, None):
            return client
        app = build_app_from_config(config)
        context = Context.from_app(
            app, uri=f"http://local-tiled-app-{username}/api/v1", api_key=api_key
        )
        contexts.append(context)
        client = from_context(context, remember_me=False)
        clients[username] = client
        if api_key is None:
            with enter_username_password(username, password):
                client.context.login(remember_me=False)
        return client

    admin_client = _create_and_login_context("admin", "admin")
    for k in ["foo", "bar", "baz", "qux"]:
        admin_client[k].write_array(arr, key="data_A", access_tags=["alice_tag"])
        admin_client[k].write_array(arr, key="data_B", access_tags=["chemists_tag"])
        admin_client[k].write_array(arr, key="data_C", access_tags=["public"])

    yield _create_and_login_context

    for context in contexts:
        context.close()


def test_access_tag_compiler(compile_access_tags_db_with_reset):
    """
    Test that compilation of access tags is working. This tests:
    - Adding and removing a tag
    - Adding and removing a role
    - Adding and removing a user from a tag
    - Adding and removing a group from a tag
    - Adding and removing from the `auto_tags` for a tag
    - Adding and removing a tag from the `tag_owners` section
    - Adding and removing users and groups from the owners of a tag
    - Adding and removing a member from a group
    - Changing a user's role/scopes on a tag
    - Changing a group's role/scopes on a tag
    - Making a tag public/not-public
    - Disallow redefining the `public` tag
    """
    access_tags_compiler = compile_access_tags_db_with_reset
    compiler_tag_config = access_tags_compiler.tag_config

    def new_group_parser(groupname):
        return {
            "chemists": ["bob", "mary", "kate"],
            "biologists": ["chris", "fred"],
            "physicists": ["sue", "tony"],
        }[groupname]

    access_tags_compiler.group_parser = new_group_parser

    compiler_tag_config["tags"].update(
        {"new_tag": {"users": [{"name": "tony", "scopes": ["read:metadata"]}]}}
    )
    compiler_tag_config["roles"].update({"new_role": {"scopes": ["read:metadata"]}})
    compiler_tag_config["tags"]["biologists_tag"]["users"].append(
        {"name": "tony", "role": "facility_user"}
    )
    compiler_tag_config["tags"]["physicists_tag"]["groups"].append(
        {"name": "biologists", "role": "facility_user"}
    )
    compiler_tag_config["tags"]["chemists_tag"]["auto_tags"].append({"name": "new_tag"})
    compiler_tag_config["tag_owners"].update({"new_tag": {"users": [{"name": "tony"}]}})
    compiler_tag_config["tag_owners"]["biologists_tag"]["users"].append(
        {"name": "tony"}
    )
    compiler_tag_config["tag_owners"]["chemists_tag"]["groups"].append(
        {"name": "biologists"}
    )
    compiler_tag_config["tags"]["alice_tag"]["users"][0]["role"] = "facility_user"
    compiler_tag_config["tags"]["biologists_tag"]["groups"][0].pop("scopes")
    compiler_tag_config["tags"]["biologists_tag"]["groups"][0].update(
        {"role": "facility_admin"}
    )
    compiler_tag_config["tags"]["alice_tag"].update({"auto_tags": [{"name": "public"}]})

    access_tags_compiler.load_tag_config()
    access_tags_compiler.recompile()

    db = sqlite3.connect("file:compiled_tags_mem?mode=memory&cache=shared", uri=True)
    cursor = db.cursor()
    # check that new tag was added and compiled with user+scopes
    cursor.execute(
        "SELECT EXISTS(SELECT 1 FROM tags WHERE name = ?);",
        ("new_tag",),
    )
    assert bool(cursor.fetchone()[0])
    cursor.execute(
        "SELECT EXISTS(SELECT 1 FROM user_tag_scopes WHERE tag_name = ? AND user_name = ?);",
        ("new_tag", "tony"),
    )
    assert bool(cursor.fetchone()[0])

    # check that new role was added - note roles do not get saved in the db
    assert "new_role" in access_tags_compiler.roles

    # check that newly added user and group were given scopes on tag
    cursor.execute(
        "SELECT EXISTS(SELECT 1 FROM user_tag_scopes WHERE tag_name = ? AND user_name = ?);",
        ("biologists_tag", "tony"),
    )
    assert bool(cursor.fetchone()[0])
    cursor.execute(
        "SELECT EXISTS(SELECT 1 FROM user_tag_scopes WHERE tag_name = ? AND user_name = ?);",
        ("physicists_tag", "chris"),
    )
    assert bool(cursor.fetchone()[0])

    # check that auto_tag added ACL to parent tag
    cursor.execute(
        "SELECT EXISTS(SELECT 1 FROM user_tag_scopes WHERE tag_name = ? AND user_name = ?);",
        ("chemists_tag", "tony"),
    )
    assert bool(cursor.fetchone()[0])

    # check tag was added to tag_owners section
    cursor.execute(
        "SELECT EXISTS(SELECT 1 FROM user_tag_owners WHERE tag_name = ?);",
        ("new_tag",),
    )
    assert bool(cursor.fetchone()[0])

    # check adding new user and group to owners of tags
    cursor.execute(
        "SELECT EXISTS(SELECT 1 FROM user_tag_owners WHERE tag_name = ? AND user_name = ?);",
        ("biologists_tag", "tony"),
    )
    assert bool(cursor.fetchone()[0])
    cursor.execute(
        "SELECT EXISTS(SELECT 1 FROM user_tag_owners WHERE tag_name = ? AND user_name = ?);",
        ("chemists_tag", "chris"),
    )
    assert bool(cursor.fetchone()[0])

    # check that the role/scopes changes for a user and group on a tag were effective
    cursor.execute(
        "SELECT NOT EXISTS(SELECT 1 FROM user_tag_scopes "
        "WHERE tag_name = ? AND user_name = ? AND scope_name = ?);",
        ("alice_tag", "alice", "write:metadata"),
    )
    assert bool(cursor.fetchone()[0])
    cursor.execute(
        "SELECT EXISTS(SELECT 1 FROM user_tag_scopes "
        "WHERE tag_name = ? AND user_name = ? AND scope_name = ?);",
        ("biologists_tag", "chris", "write:metadata"),
    )
    assert bool(cursor.fetchone()[0])

    # check tha tag was marked as public after inheriting public tag
    cursor.execute(
        "SELECT EXISTS(SELECT 1 FROM public_tags WHERE name = ?);",
        ("alice_tag",),
    )
    assert bool(cursor.fetchone()[0])

    # check that user added to group was compiled into tag ACL
    cursor.execute(
        "SELECT EXISTS(SELECT 1 FROM user_tag_scopes WHERE tag_name = ? AND user_name = ?);",
        ("chemists_tag", "kate"),
    )
    assert bool(cursor.fetchone()[0])

    # attempt redefining the public tag (and fail)
    compiler_tag_config["tags"].update(
        {"public": {"users": [{"name": "tony", "scopes": ["read:metadata"]}]}}
    )
    access_tags_compiler.load_tag_config()
    with pytest.raises(ValueError):
        access_tags_compiler.recompile()

    # remove all changes/additions by reverting to the original config
    access_tags_compiler.tag_config = access_tag_config
    access_tags_compiler.group_parser = group_parser
    access_tags_compiler.clear_raw_tags()
    access_tags_compiler.load_tag_config()
    access_tags_compiler.recompile()

    # check that new tag was removed and no longer compiled
    cursor.execute(
        "SELECT NOT EXISTS(SELECT 1 FROM tags WHERE name = ?);",
        ("new_tag",),
    )
    assert bool(cursor.fetchone()[0])
    cursor.execute(
        "SELECT NOT EXISTS(SELECT 1 FROM user_tag_scopes WHERE tag_name = ? AND user_name = ?);",
        ("new_tag", "tony"),
    )
    assert bool(cursor.fetchone()[0])

    # check that new role was removed - note roles do not get saved in the db
    assert "new_role" not in access_tags_compiler.roles

    # check that removed user and group were not given scopes on tag
    cursor.execute(
        "SELECT NOT EXISTS(SELECT 1 FROM user_tag_scopes WHERE tag_name = ? AND user_name = ?);",
        ("biologists_tag", "tony"),
    )
    assert bool(cursor.fetchone()[0])
    cursor.execute(
        "SELECT NOT EXISTS(SELECT 1 FROM user_tag_scopes WHERE tag_name = ? AND user_name = ?);",
        ("physicists_tag", "chris"),
    )
    assert bool(cursor.fetchone()[0])

    # check that auto_tag ACL removed from parent tag
    cursor.execute(
        "SELECT NOT EXISTS(SELECT 1 FROM user_tag_scopes WHERE tag_name = ? AND user_name = ?);",
        ("chemists_tag", "tony"),
    )
    assert bool(cursor.fetchone()[0])

    # check tag was removed from tag_owners section
    cursor.execute(
        "SELECT NOT EXISTS(SELECT 1 FROM user_tag_owners WHERE tag_name = ?);",
        ("new_tag",),
    )
    assert bool(cursor.fetchone()[0])

    # check removing user and group from owners of tags
    cursor.execute(
        "SELECT NOT EXISTS(SELECT 1 FROM user_tag_owners WHERE tag_name = ? AND user_name = ?);",
        ("biologists_tag", "tony"),
    )
    assert bool(cursor.fetchone()[0])
    cursor.execute(
        "SELECT NOT EXISTS(SELECT 1 FROM user_tag_owners WHERE tag_name = ? AND user_name = ?);",
        ("chemists_tag", "chris"),
    )
    assert bool(cursor.fetchone()[0])

    # check that the role/scopes changes for a user and group on a tag were undone
    cursor.execute(
        "SELECT EXISTS(SELECT 1 FROM user_tag_scopes "
        "WHERE tag_name = ? AND user_name = ? AND scope_name = ?);",
        ("alice_tag", "alice", "write:metadata"),
    )
    assert bool(cursor.fetchone()[0])
    cursor.execute(
        "SELECT NOT EXISTS(SELECT 1 FROM user_tag_scopes "
        "WHERE tag_name = ? AND user_name = ? AND scope_name = ?);",
        ("biologists_tag", "chris", "write:metadata"),
    )
    assert bool(cursor.fetchone()[0])

    # check tha tag was unmarked as public after removing the public auto_tag
    cursor.execute(
        "SELECT NOT EXISTS(SELECT 1 FROM public_tags WHERE name = ?);",
        ("alice_tag",),
    )
    assert bool(cursor.fetchone()[0])

    # check that user removed from group was compiled out of tag ACL
    cursor.execute(
        "SELECT NOT EXISTS(SELECT 1 FROM user_tag_scopes WHERE tag_name = ? AND user_name = ?);",
        ("chemists_tag", "kate"),
    )
    assert bool(cursor.fetchone()[0])


def test_basic_access_control(access_control_test_context_factory):
    """
    Test that basic access control and tag compilation are working.
    Only tests simple visibility of the data (i.e. "read:metadata" scope),
      does not tests writing or full reading of the data.

    In other words, tests that compiled tags allow/disallow access including:
      - top-level tags
      - tags directly on datasets
      - tags "inherited" on datasets (auto_tags)
      - "public" tags on datasets
      - groups compiled into tags
      - scopes compiled into tags by a role
      - scopes compiled into tags by a scopes list
      - nested access blocked by upper tags (even if deeper tags would permit access)

    Note: MapAdapter does not support access control. As such, the server root
          does not currently filter top-level entries.
    """
    alice_client = access_control_test_context_factory("alice", "alice")
    bob_client = access_control_test_context_factory("bob", "bob")

    top = "foo"
    assert top in alice_client
    # no access control on MapAdapter - can't filter top-level yet
    # assert top not in bob_client
    for data in ["data_A", "data_B", "data_C"]:
        # Alice has access below the top-level, given by a direct tag
        # Bob does not have access to any data, blocked by the top-level's tag
        # data_A - alice has access given by a direct tag of which they are a user
        # data_B - alice has access given by an inherited tag
        # data_C - alice has access given by a public tag
        assert data in alice_client[top]
        alice_client[top][data]
        with pytest.raises(KeyError):
            bob_client[top][data]

    top = "bar"
    assert top in alice_client
    assert top in bob_client
    for data in ["data_A"]:
        # Alice has access below the top-level, given by an inherited tag
        # data_A - bob does not have access conferred by any tags
        assert data in alice_client[top]
        alice_client[top][data]
        assert data not in bob_client[top]
        with pytest.raises(KeyError):
            bob_client[top][data]
    for data in ["data_B", "data_C"]:
        # Bob has access below the top-level, given by a direct tag of which they are in a group
        # data_B - alice has scopes compiled in via role
        # data_B - bob has access given by a direct tag of which they are in a group
        # data_B - bob has scopes compiled in via list of scopes
        # data_C - alice and bob are given access by a public tag
        assert data in alice_client[top]
        alice_client[top][data]
        assert data in bob_client[top]
        bob_client[top][data]


def test_writing_access_control(access_control_test_context_factory):
    """
    Test that writing access control and tag ownership is working.
    Only tests that the writing request does not fail.
    Does not test the written data for validity.

    This tests the following:
      - Writing without applying an access tag
      - Writing while applying an access tag the user owns
      - Writing while applying an access tag the user does not own
      - Writing while applying an access tag that is not defined
      - Writing while applying the "public" tag (admin only)
      - Writing into a location where the user does not have write access
      - Writing while applying an access tag the user owns through group membership
      - Writing while applying multiple access tags
      - Writing while applying a tag which does not give the user the minimum scopes
    """

    alice_client = access_control_test_context_factory("alice", "alice")
    bob_client = access_control_test_context_factory("bob", "bob")
    sue_client = access_control_test_context_factory("sue", "sue")

    top = "foo"
    alice_client[top].write_array(arr, key="data_Q")
    alice_client[top].write_array(arr, key="data_R", access_tags=["alice_tag"])
    with fail_with_status_code(HTTP_403_FORBIDDEN):
        alice_client[top].write_array(arr, key="data_S", access_tags=["chemists_tag"])
    with fail_with_status_code(HTTP_403_FORBIDDEN):
        alice_client[top].write_array(arr, key="data_T", access_tags=["undefined_tag"])
    with fail_with_status_code(HTTP_403_FORBIDDEN):
        alice_client[top].write_array(arr, key="data_U", access_tags=["public"])

    top = "bar"
    with fail_with_status_code(HTTP_403_FORBIDDEN):
        bob_client[top].write_array(arr, key="data_V")

    top = "baz"
    sue_client[top].write_array(
        arr, key="data_W", access_tags=["physicists_tag", "chemists_tag"]
    )
    access_tags = sue_client[top]["data_W"].access_blob["tags"]
    assert "physicists_tag" in access_tags
    assert "chemists_tag" in access_tags
    with fail_with_status_code(HTTP_403_FORBIDDEN):
        sue_client[top].write_array(arr, key="data_X", access_tags=["chemists_tag"])


def test_user_owned_node_access_control(access_control_test_context_factory):
    """
    Test that user-owned nodes (i.e. nodes created without access tags applied)
      are visible after creation and can be modified by the user.
    Also test that the data is visible after a tag is applied, and
      that other users cannot see user-owned nodes.
    """

    alice_client = access_control_test_context_factory("alice", "alice")
    bob_client = access_control_test_context_factory("bob", "bob")

    top = "foo"
    for data in ["data_Y"]:
        # Create a new user-owned node
        alice_client[top].write_array(arr, key=data)
        assert data in alice_client[top]
        alice_client[top][data]
        access_blob = alice_client[top][data].access_blob
        assert "user" in access_blob
        assert "alice" in access_blob["user"]
        # Convert from user-owned node to a tagged node
        alice_client[top][data].replace_metadata(access_tags=["alice_tag"])
        access_blob = alice_client[top][data].access_blob
        assert "user" not in access_blob
        assert "tags" in access_blob
        assert "alice_tag" in access_blob["tags"]
        assert data in alice_client[top]
        alice_client[top][data]

    top = "bar"
    for data in ["data_Z"]:
        # Create a user-owned node and check that it is access restricted
        alice_client[top].write_array(arr, key=data)
        assert data not in bob_client[top]
        with pytest.raises(KeyError):
            bob_client[top][data]


def test_public_anonymous_access_control(access_control_test_context_factory):
    """
    Test that data which is tagged public is visible to unauthenticated
      (anonymous) users when the server allows anonymous access.
    """
    zoe_client = access_control_test_context_factory("zoe", "zoe")
    zoe_client.logout()
    anon_client = zoe_client

    top = "qux"
    assert top in anon_client
    for data in ["data_A", "data_B"]:
        assert data not in anon_client[top]
        with pytest.raises(KeyError):
            anon_client[top][data]
    for data in ["data_C"]:
        assert data in anon_client[top]
        anon_client[top][data]


def test_admin_access_control(access_control_test_context_factory):
    """
    Test that admin accounts have various elevated privileges, including:
    - Apply/remove public tag to/from a node
    - Apply/remove tags while ignoring minimum scopes
    - Apply/remove tags that the user does not own
    - View all data regardless of tags
    - Apply an access tag that is not defined (disallowed)
    - Remove all tags from a node, but still view that node
    - Also includes test of an empty tags list blocking access for regular users
    """
    admin_client = access_control_test_context_factory("admin", "admin")
    alice_client = access_control_test_context_factory("alice", "alice")

    top = "foo"
    for data in ["data_L"]:
        # create a node and tag it public
        admin_client[top].write_array(arr, key=data, access_tags=["public"])
        assert data in alice_client[top]
        alice_client[top][data]
        # remove public access, in fact remove all tags and ignore missing scopes
        admin_client[top][data].replace_metadata(access_tags=[])
        assert data in admin_client[top]
        admin_client[top][data]
        assert data not in alice_client[top]
        with pytest.raises(KeyError):
            alice_client[top][data]
        # apply a tag that the admin user does not own and ignore missing scopes
        admin_client[top][data].replace_metadata(access_tags=["chemists_tag"])
        assert data in admin_client[top]
        admin_client[top][data]
        assert data in alice_client[top]
        alice_client[top][data]
        # remove a tag that the admin user does not own
        admin_client[top][data].replace_metadata(access_tags=["chemists_tag"])
        # apply a tag which is not defined
        with fail_with_status_code(HTTP_403_FORBIDDEN):
            admin_client[top][data].replace_metadata(access_tags=["undefined_tag"])


def test_update_node_access_control(access_control_test_context_factory):
    """
    Test that access control on metadata changes is working.

    This tests the following:
      - Update metadata while having write access
      - Prevent updating metadata without having write access
      - Successfully add an access tag and remove an access tag
      - Prevent adding or removing an access tag without having write access
      - Prevent adding or removing access tags which the user does not own
      - Add and remove access tags which do not confer the necessary scopes
      - Attempt to add an undefined access tag (not allowed)
      - Attempt to add the "public" tag (admin only)
      - Attempt to remove the "public" tag (admin only)
      - Attempt to remove an undefined access tag (not allowed)
    """
    admin_client = access_control_test_context_factory("admin", "admin")
    alice_client = access_control_test_context_factory("alice", "alice")
    chris_client = access_control_test_context_factory("chris", "chris")
    sue_client = access_control_test_context_factory("sue", "sue")

    top = "qux"
    for data in ["data_F"]:
        admin_client[top].write_array(arr, key=data, access_tags=["alice_tag"])
        # successfully update metadata, user has write access
        alice_client[top][data].replace_metadata(metadata={"materials": ["Cu", "Ag"]})
        assert "Ag" in alice_client[top][data].metadata["materials"]
        # fail to update metadata, user does not have write access
        with fail_with_status_code(HTTP_403_FORBIDDEN):
            chris_client[top][data].replace_metadata(
                metadata={"materials": ["Ag", "Au"]}
            )
        assert "Au" not in chris_client[top][data].metadata["materials"]

        # succeeds to add a new access tag and remove the old access tag
        alice_client[top][data].replace_metadata(access_tags=["biologists_tag"])
        access_tags = alice_client[top][data].access_blob["tags"]
        assert "alice_tag" not in access_tags
        assert "biologists_tag" in access_tags

        # fails to add a new access tag, user does not have write access
        with fail_with_status_code(HTTP_403_FORBIDDEN):
            chris_client[top][data].replace_metadata(
                access_tags=["biologists_tag", "chris_tag"]
            )
        admin_client[top][data].replace_metadata(
            access_tags=["alice_tag", "biologists_tag"]
        )
        # fails to remove an access tag, user does not have write access
        with fail_with_status_code(HTTP_403_FORBIDDEN):
            chris_client[top][data].replace_metadata(access_tags=["biologists_tag"])
        admin_client[top][data].replace_metadata(access_tags=["biologists_tag"])

        # fails to add a new access tag, user does not own the tag
        with fail_with_status_code(HTTP_403_FORBIDDEN):
            alice_client[top][data].replace_metadata(
                access_tags=["biologists_tag", "chris_tag"]
            )
        admin_client[top][data].replace_metadata(
            access_tags=["biologists_tag", "chris_tag"]
        )
        # fails to remove an access tag, user does not own the tag
        with fail_with_status_code(HTTP_403_FORBIDDEN):
            chris_client[top][data].replace_metadata(access_tags=["biologists_tag"])
        admin_client[top][data].replace_metadata(access_tags=["biologists_tag"])

        # fail to add an undefined tag
        with fail_with_status_code(HTTP_403_FORBIDDEN):
            alice_client[top][data].replace_metadata(
                access_tags=["undefined_tag", "biologists_tag"]
            )
        # fail to add the "public" tag
        with fail_with_status_code(HTTP_403_FORBIDDEN):
            alice_client[top][data].replace_metadata(
                access_tags=["public", "biologists_tag"]
            )
        admin_client[top][data].replace_metadata(
            access_tags=["public", "biologists_tag"]
        )
        # fail to remove the "public" tag
        with fail_with_status_code(HTTP_403_FORBIDDEN):
            alice_client[top][data].replace_metadata(access_tags=["biologists_tag"])
        admin_client[top][data].replace_metadata(access_tags=["biologists_tag"])

        # surgically add an undefined tag to the node, then fail when trying to remove it
        db = sqlite3.connect(f"file:catalog_{top}?mode=memory&cache=shared", uri=True)
        cursor = db.cursor()
        cursor.execute(
            "UPDATE nodes SET access_blob = ? WHERE key = ?",
            ('{"tags": ["undefined_tag", "biologists_tag"]}', data),
        )
        db.commit()
        with fail_with_status_code(HTTP_403_FORBIDDEN):
            alice_client[top][data].replace_metadata(access_tags=["biologists_tag"])

    top = "baz"
    for data in ["data_G"]:
        sue_client[top].write_array(arr, key=data)
        # fail to apply a new access tag as it does not give the user the
        # minimum required scopes
        # this case only affects user-owned nodes:
        # - if we did not have read access, we would not even see the node
        # - if we did not have write access, we would be blocked by scopes
        # - if an existing tag already gave us read and write, adding a tag would succeed
        # - if an existing tag already gave us read and write, and we tried to remove it
        #   while adding the new tag, it's really the removal operation that prevents this,
        #   and this operation is tested below
        # this leaves only user-owned nodes (full access for the user, but no existing tags)
        with fail_with_status_code(HTTP_403_FORBIDDEN):
            sue_client[top][data].replace_metadata(access_tags=["chemists_tag"])
        sue_client[top][data].replace_metadata(
            access_tags=["physicists_tag", "chemists_tag"]
        )
        # fail to apply the new access tag as removing the old access tag results
        # in insufficent scopes for the user
        with fail_with_status_code(HTTP_403_FORBIDDEN):
            sue_client[top][data].replace_metadata(access_tags=["chemists_tag"])


def test_empty_access_blob_access_control(access_control_test_context_factory):
    """
    Test the cases where a node in the catalog has an empty access blob.
    This case occurs when migrating an older catalog without also
      populating the access_blob column.
    """
    admin_client = access_control_test_context_factory("admin", "admin")
    alice_client = access_control_test_context_factory("alice", "alice")

    top = "qux"
    for data in ["data_M"]:
        admin_client[top].write_array(arr, key=data, access_tags=["alice_tag"])
        db = sqlite3.connect(f"file:catalog_{top}?mode=memory&cache=shared", uri=True)
        cursor = db.cursor()
        cursor.execute(
            "UPDATE nodes SET access_blob = json('{}') WHERE key == 'data_M'"
        )
        db.commit()

        assert data in admin_client[top]
        admin_client[top][data]
        assert data not in alice_client[top]
        with pytest.raises(KeyError):
            alice_client[top][data]


def test_container_access_control(access_control_test_context_factory):
    """
    Test that access control for data nested in containers allows/denies access.
    This mostly checks that if a user does not have access to a container,
      that user cannot reach inside the container to view data they would
      otherwise have access for.
    """
    alice_client = access_control_test_context_factory("alice", "alice")
    sue_client = access_control_test_context_factory("sue", "sue")

    top = "baz"
    for c in ["C1"]:
        alice_client[top].create_container(key=c, access_tags=["alice_tag"])
        alice_client[top][c].write_array(
            arr, key=f"{c}_array", access_tags=["physicists_tag"]
        )
        assert f"{c}_array" in alice_client[top][c]
        alice_client[top][c][f"{c}_array"]
        assert c not in sue_client[top]
        with pytest.raises(KeyError):
            sue_client[top][c][f"{c}_array"]


def test_node_export_access_control(
    access_control_test_context_factory, buffer_factory
):
    """
    Test access control when exporting from Tiled (here: a container).
    These tests include:
    - Test that top-level nodes are disincluded appropriately
      (MapAdapter->CatalogAdapter transition).
    - Test that basic export works - i.e. nodes for which the user has
      access are included.
    - Test that nodes for which the user does not have access are not included.
    - Test that this behavior also works for user-owned (untagged) nodes.
    """
    alice_client = access_control_test_context_factory("alice", "alice")
    sue_client = access_control_test_context_factory("sue", "sue")

    top = "baz"
    alice_client[top].write_array(arr, key="data_D")
    sue_client[top].write_array(arr, key="data_E")

    alice_export_buffer = buffer_factory()
    sue_export_buffer = buffer_factory()

    alice_client.export(alice_export_buffer, format="application/json")
    sue_client.export(sue_export_buffer, format="application/json")

    alice_export_buffer.seek(0)
    sue_export_buffer.seek(0)

    alice_exported_data = json.loads(alice_export_buffer.read())
    sue_exported_data = json.loads(sue_export_buffer.read())

    top = "foo"
    assert top in alice_exported_data["contents"]
    assert top not in sue_exported_data["contents"]
    for data in ["data_A", "data_B", "data_C"]:
        assert data in alice_exported_data["contents"][top]["contents"]
        alice_exported_data["contents"][top]["contents"][data]

    top = "baz"
    assert top in alice_exported_data["contents"]
    assert top in sue_exported_data["contents"]
    for data in ["data_A", "data_B", "data_D"]:
        assert data not in sue_exported_data["contents"][top]["contents"]
        with pytest.raises(KeyError):
            sue_exported_data["contents"][top]["contents"][data]
    for data in ["data_C", "data_E"]:
        assert data in sue_exported_data["contents"][top]["contents"]
        sue_exported_data["contents"][top]["contents"][data]


def test_apikey_auth_access_control(access_control_test_context_factory):
    """
    Test access control when authenticated by an API key, including:
    - Allow basic access with an API key that is not tag-restricted
    - Disallow access to tags that are not added to a tag-restricted API key
    - Allow access to tags that are added to a tag-restricted API key
    - User-owned node access/writing is blocked when using a tag-restricted API key
    """
    alice_client = access_control_test_context_factory("alice", "alice")
    alice_apikey_info = alice_client.context.create_api_key()
    alice_client.logout()
    alice_client.context.api_key = alice_apikey_info["secret"]

    top = "foo"
    for data in ["data_A"]:
        assert data in alice_client[top]
        alice_client[top][data]

    top = "bar"
    alice_client[top].write_array(arr, key="data_O")

    alice_apikey_info = alice_client.context.create_api_key(
        access_tags=["chemists_tag"]
    )
    alice_client.context.api_key = alice_apikey_info["secret"]

    top = "bar"
    for data in ["data_A"]:
        assert data not in alice_client[top]
        with pytest.raises(KeyError):
            alice_client[top][data]
    for data in ["data_B"]:
        assert data in alice_client[top]
        alice_client[top][data]
    for data in ["data_O"]:
        assert data not in alice_client[top]
        with pytest.raises(KeyError):
            alice_client[top][data]
    with fail_with_status_code(HTTP_403_FORBIDDEN):
        alice_client[top].write_array(arr, key="data_P")


def test_service_principal_access_control(
    access_control_test_context_factory, compile_access_tags_db_with_reset
):
    """
    Test that access control works for service principals.
    Creates a service principal and updates the access tag config to
      add this prinicpal to a tag.
    """
    admin_client = access_control_test_context_factory("admin", "admin")
    sp = admin_client.context.admin.create_service_principal("user")
    sp_apikey_info = admin_client.context.admin.create_api_key(sp["uuid"])
    sp_client = access_control_test_context_factory(
        sp["uuid"], api_key=sp_apikey_info["secret"]
    )

    access_tags_compiler = compile_access_tags_db_with_reset
    compiler_tag_config = access_tags_compiler.tag_config

    compiler_tag_config["tags"]["physicists_tag"]["users"].append(
        {"name": sp["uuid"], "role": "facility_admin"}
    )
    compiler_tag_config["tag_owners"]["physicists_tag"]["users"].append(
        {"name": sp["uuid"]}
    )

    access_tags_compiler.load_tag_config()
    access_tags_compiler.recompile()

    top = "baz"
    for data in ["data_A"]:
        assert data not in sp_client[top]
        with pytest.raises(KeyError):
            sp_client[top][data]
    for data in ["data_N"]:
        sp_client[top].write_array(arr, key=data, access_tags=["physicists_tag"])
        assert data in sp_client[top]
        sp_client[top][data]
